"""Type definitions for Zabbix API objects.

Since we are supporting multiple versions of the Zabbix API at the same time,
we don't operate with very strict type definitions. Some definitions are
TypedDicts, while others are Pydantic models. All models are able to
take extra fields, since we don't know (or always care) which fields are
present in which API versions.

It's not type-safe, but it's better than nothing. In the future, we might
want to look into defining Pydantic models that can accommodate multiple
Zabbix versions.
"""

from __future__ import annotations

import logging
from collections.abc import Iterable
from collections.abc import MutableMapping
from collections.abc import Sequence
from datetime import datetime
from datetime import timedelta
from typing import Annotated
from typing import Any
from typing import Optional
from typing import Union

from pydantic import AliasChoices
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import FieldSerializationInfo
from pydantic import PlainSerializer
from pydantic import SerializationInfo
from pydantic import ValidationError
from pydantic import ValidationInfo
from pydantic import ValidatorFunctionWrapHandler
from pydantic import WrapValidator
from pydantic import computed_field
from pydantic import field_serializer
from pydantic import field_validator
from pydantic_core import PydanticCustomError
from typing_extensions import Literal
from typing_extensions import TypeAliasType
from typing_extensions import TypedDict

from zabbix_cli.models import ColsRowsType
from zabbix_cli.models import RowsType
from zabbix_cli.models import TableRenderable
from zabbix_cli.output.style import Color
from zabbix_cli.pyzabbix.enums import AckStatus
from zabbix_cli.pyzabbix.enums import ActiveInterface
from zabbix_cli.pyzabbix.enums import EventStatus
from zabbix_cli.pyzabbix.enums import GUIAccess
from zabbix_cli.pyzabbix.enums import HostgroupFlag
from zabbix_cli.pyzabbix.enums import HostgroupType
from zabbix_cli.pyzabbix.enums import InterfaceConnectionMode
from zabbix_cli.pyzabbix.enums import InterfaceType
from zabbix_cli.pyzabbix.enums import ItemType
from zabbix_cli.pyzabbix.enums import MacroType
from zabbix_cli.pyzabbix.enums import MaintenancePeriodType
from zabbix_cli.pyzabbix.enums import MaintenanceStatus
from zabbix_cli.pyzabbix.enums import MaintenanceWeekType
from zabbix_cli.pyzabbix.enums import MediaTypeType
from zabbix_cli.pyzabbix.enums import MonitoredBy
from zabbix_cli.pyzabbix.enums import MonitoringStatus
from zabbix_cli.pyzabbix.enums import ProxyCompatibility
from zabbix_cli.pyzabbix.enums import ProxyGroupState
from zabbix_cli.pyzabbix.enums import ProxyMode
from zabbix_cli.pyzabbix.enums import ProxyModePre70
from zabbix_cli.pyzabbix.enums import TriggerPriority
from zabbix_cli.pyzabbix.enums import UsergroupPermission
from zabbix_cli.pyzabbix.enums import UsergroupStatus
from zabbix_cli.pyzabbix.enums import UserRole
from zabbix_cli.pyzabbix.enums import ValueType
from zabbix_cli.utils.utils import get_maintenance_active_days
from zabbix_cli.utils.utils import get_maintenance_active_months
from zabbix_cli.utils.utils import get_maintenance_status
from zabbix_cli.utils.utils import get_monitoring_status

logger = logging.getLogger(__name__)

SortOrder = Literal["ASC", "DESC"]


# Source: https://docs.pydantic.dev/2.7/concepts/types/#named-recursive-types
def json_custom_error_validator(
    value: Any, handler: ValidatorFunctionWrapHandler, _info: ValidationInfo
) -> Any:
    """Simplify the error message to avoid a gross error stemming from
    exhaustive checking of all union options.
    """  # noqa: D205
    try:
        return handler(value)
    except ValidationError:
        raise PydanticCustomError(
            "invalid_json",
            "Input is not valid json",
        ) from None


Json = TypeAliasType(
    "Json",
    Annotated[
        Union[
            MutableMapping[str, "Json"],
            Sequence["Json"],
            str,
            int,
            float,
            bool,
            None,
        ],
        WrapValidator(json_custom_error_validator),
    ],
)


ParamsType = MutableMapping[str, Json]
"""Type used to construct parameters for API requests.
Can only contain native JSON-serializable types.

BaseModel objects must be converted to JSON-serializable dicts before being
assigned as values in a ParamsType.
"""


def serialize_host_list_json(
    hosts: list[Host], info: SerializationInfo
) -> list[dict[str, str]]:
    """Custom JSON serializer for a list of hosts.

    Most of the time we don't want to serialize _all_ fields of a host.
    This serializer assumes that we want to serialize the minimal representation
    of hosts unless the context specifies otherwise.
    """
    if isinstance(info.context, dict):
        if info.context.get("full_host"):  # pyright: ignore[reportUnknownMemberType]
            return [host.model_dump(mode="json") for host in hosts]
    return [host.model_simple_dump() for host in hosts]


HostList = Annotated[
    list["Host"], PlainSerializer(serialize_host_list_json, when_used="json")
]
"""List of hosts that serialize as the minimal representation of a list of hosts."""


def age_from_datetime(dt: Optional[datetime]) -> Optional[str]:
    """Returns the age of a datetime object as a human-readable
    string, or None if the datetime is None.
    """
    if not dt:
        return None
    n = datetime.now(tz=dt.tzinfo)
    age = n - dt
    # strip microseconds
    return str(age - timedelta(microseconds=age.microseconds))


def format_datetime(dt: Optional[datetime]) -> str:
    """Returns a formatted datetime string, or empty string if the
    datetime is None.
    """
    if not dt:
        return ""
    return dt.strftime("%Y-%m-%d %H:%M:%S")


class ModifyHostItem(TypedDict):
    """Argument for a host ID in an API request."""

    hostid: Union[str, int]


ModifyHostParams = list[ModifyHostItem]

"""A list of host IDs in an API request.

E.g. `[{"hostid": "123"}, {"hostid": "456"}]`
"""


class ModifyGroupItem(TypedDict):
    """Argument for a group ID in an API request."""

    groupid: Union[str, int]


ModifyGroupParams = list[ModifyGroupItem]
"""A list of host/template group IDs in an API request.

E.g. `[{"groupid": "123"}, {"groupid": "456"}]`
"""


class ModifyTemplateItem(TypedDict):
    """Argument for a template ID in an API request."""

    templateid: Union[str, int]


ModifyTemplateParams = list[ModifyTemplateItem]
"""A list of template IDs in an API request.

E.g. `[{"templateid": "123"}, {"templateid": "456"}]`
"""


class ZabbixAPIError(BaseModel):
    """Zabbix API error information."""

    code: int
    message: str
    data: Optional[str] = None


class ZabbixAPIResponse(BaseModel):
    """The raw response from the Zabbix API"""

    jsonrpc: str
    id: int
    result: Any = None  # can subclass this and specify types (ie. ZabbixAPIListResponse, ZabbixAPIStrResponse, etc.)
    """Result of API call, if request succeeded."""
    error: Optional[ZabbixAPIError] = None
    """Error info, if request failed."""


class ZabbixAPIBaseModel(TableRenderable):
    """Base model for Zabbix API objects.

    Implements the `TableRenderable` interface, which allows us to render
    it as a table, JSON, csv, etc.
    """

    model_config = ConfigDict(validate_assignment=True, extra="ignore")

    def model_dump_api(self) -> dict[str, Any]:
        """Dump the model as a JSON-serializable dict used in API calls.

        Excludes computed fields by default.
        """
        return self.model_dump(
            mode="json",
            exclude=set(self.model_computed_fields),
            exclude_none=True,
        )


class ZabbixRight(ZabbixAPIBaseModel):
    permission: int
    id: str
    name: Optional[str] = None  # name of group (injected by application)

    @computed_field
    @property
    def permission_str(self) -> str:
        """Returns the permission as a formatted string."""
        return UsergroupPermission.string_from_value(self.permission)

    def model_dump_api(self) -> dict[str, Any]:
        return self.model_dump(
            mode="json", include={"permission", "id"}, exclude_none=True
        )


class User(ZabbixAPIBaseModel):
    userid: str
    username: str = Field(..., validation_alias=AliasChoices("username", "alias"))
    name: Optional[str] = None
    surname: Optional[str] = None
    url: Optional[str] = None
    autologin: Optional[str] = None
    autologout: Optional[str] = None
    roleid: Optional[int] = Field(
        default=None, validation_alias=AliasChoices("roleid", "type")
    )
    # NOTE: Not adding properties we don't use, since Zabbix has a habit of breaking
    # its own API by changing names and types of properties between versions.

    @computed_field
    @property
    def role(self) -> Optional[str]:
        """Returns the role name, if available."""
        if self.roleid:
            return UserRole.string_from_value(self.roleid)
        return None

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["UserID", "Username", "Name", "Surname", "Role"]
        rows: RowsType = [
            [
                self.userid,
                self.username,
                self.name or "",
                self.surname or "",
                self.role or "",
            ]
        ]
        return cols, rows


class Usergroup(ZabbixAPIBaseModel):
    name: str
    usrgrpid: str
    gui_access: int
    users_status: int
    rights: list[ZabbixRight] = []
    hostgroup_rights: list[ZabbixRight] = []
    templategroup_rights: list[ZabbixRight] = []
    users: list[User] = []

    @computed_field
    @property
    def gui_access_str(self) -> str:
        """GUI access code as a formatted string."""
        return GUIAccess.string_from_value(self.gui_access)

    @computed_field
    @property
    def users_status_str(self) -> str:
        """User status as a formatted string."""
        return UsergroupStatus.string_from_value(self.users_status)

    # LEGACY
    @computed_field
    @property
    def status(self) -> str:
        """LEGACY: 'users_status' is called 'status' in V2.
        Ensures serialized output contains the field.
        """
        return UsergroupStatus.string_from_value(self.users_status, with_code=True)

    # LEGACY
    @field_serializer("gui_access")
    def _LEGACY_type_serializer(
        self, v: Optional[int], _info: FieldSerializationInfo
    ) -> Union[str, int, None]:
        """Serializes the GUI access status as a formatted string in legacy JSON mode"""
        if self.legacy_json_format:
            return GUIAccess.string_from_value(v, with_code=True)
        return v


class Template(ZabbixAPIBaseModel):
    """A template object. Can contain hosts and other templates."""

    templateid: str
    host: str
    hosts: HostList = []
    macros: list[Macro] = []

    templates: list[Template] = []
    """Child templates (templates inherited from this template)."""

    parent_templates: list[Template] = Field(
        default_factory=list,
        validation_alias=AliasChoices("parentTemplates", "parent_templates"),
        serialization_alias="parentTemplates",  # match JSON output to API format
    )
    """Parent templates (templates this template inherits from)."""

    name: Optional[str] = None
    """The visible name of the template."""

    def __str__(self) -> str:
        return f"{self.name or self.host!r} ({self.templateid})"

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["ID", "Name", "Hosts", "Parents", "Children"]
        rows: RowsType = [
            [
                self.templateid,
                self.name or self.host,  # prefer name, fall back on host
                "\n".join([host.host for host in self.hosts]),
                "\n".join([parent.host for parent in self.parent_templates]),
                "\n".join([template.host for template in self.templates]),
            ]
        ]
        return cols, rows


class TemplateGroup(ZabbixAPIBaseModel):
    groupid: str
    name: str
    uuid: str
    templates: list[Template] = []


class HostGroup(ZabbixAPIBaseModel):
    groupid: str
    name: str
    hosts: HostList = []
    flags: int = 0
    internal: Optional[int] = None  # <6.2
    templates: list[Template] = []  # <6.2

    def __cols_rows__(self) -> ColsRowsType:
        # FIXME: is this ever used? Can we remove?
        cols = ["GroupID", "Name", "Flag", "Type", "Hosts"]
        rows: RowsType = [
            [
                self.groupid,
                self.name,
                HostgroupFlag.string_from_value(self.flags),
                HostgroupType.string_from_value(self.internal),
                ", ".join([host.host for host in self.hosts]),
            ]
        ]
        return cols, rows


class DictModel(ZabbixAPIBaseModel):
    """An adapter for a dict that allows it to be rendered as a table."""

    model_config = ConfigDict(extra="allow")

    def items(self) -> Iterable[tuple[str, Any]]:
        return self.model_dump(mode="python").items()

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["Key", "Value"]
        rows: RowsType = [[key, str(value)] for key, value in self.items() if value]
        return cols, rows


# TODO: expand Host model with all possible fields
# Add alternative constructor to construct from API result
class Host(ZabbixAPIBaseModel):
    hostid: str
    host: str = ""
    description: Optional[str] = None
    groups: list[HostGroup] = Field(
        default_factory=list,
        # Compat for >= 6.2.0
        validation_alias=AliasChoices("groups", "hostgroups"),
    )
    templates: list[Template] = Field(
        default_factory=list,
        validation_alias=AliasChoices("templates", "parentTemplates"),
    )
    inventory: DictModel = Field(default_factory=DictModel)
    monitored_by: Optional[MonitoredBy] = None
    proxyid: Optional[str] = Field(
        default=None,
        # Compat for <7.0.0
        validation_alias=AliasChoices("proxyid", "proxy_hostid"),
    )
    proxy_groupid: Optional[str] = None  # >= 7.0
    maintenance_status: Optional[str] = None
    # active_available is a new field in 7.0.
    # Previous versions required checking the `available` field of its first interface.
    # In zabbix-cli v2, this value was serialized as `zabbix_agent`.
    active_available: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices(
            "active_available",  # >= 7.0
            "zabbix_agent",  # Zabbix-cli V2 name of this field
        ),
    )
    status: Optional[str] = None
    macros: list[Macro] = Field(default_factory=list)
    interfaces: list[HostInterface] = Field(default_factory=list)

    # HACK: Add a field for the host's proxy that we can inject later
    proxy: Optional[Proxy] = None

    def __str__(self) -> str:
        return f"{self.host!r} ({self.hostid})"

    def model_simple_dump(self) -> dict[str, str]:
        """Dump the model with minimal fields for simple output."""
        return {
            "host": self.host,
            "hostid": self.hostid,
        }

    def set_proxy(self, proxy_map: dict[str, Proxy]) -> None:
        """Set proxy info for the host given a mapping of proxy IDs to proxies."""
        if not (proxy := proxy_map.get(str(self.proxyid))):
            return
        self.proxy = proxy

    def get_active_status(self, *, with_code: bool = False) -> str:
        """Returns the active interface status as a formatted string."""
        if self.zabbix_version.release >= (7, 0, 0):
            return ActiveInterface.string_from_value(
                self.active_available, with_code=with_code
            )
        # We are on pre-7.0.0, check the first interface
        iface = self.interfaces[0] if self.interfaces else None
        if iface:
            return ActiveInterface.string_from_value(
                iface.available, with_code=with_code
            )
        else:
            return ActiveInterface.UNKNOWN.as_status(with_code=with_code)

    # Legacy V2 JSON format compatibility
    @field_serializer("maintenance_status", when_used="json")
    def _LEGACY_maintenance_status_serializer(
        self, v: Optional[str], _info: FieldSerializationInfo
    ) -> Optional[str]:
        """Serializes the maintenance status as a formatted string
        in legacy mode, and as-is in new mode.
        """
        if self.legacy_json_format:
            return get_maintenance_status(v, with_code=True)
        return v

    @computed_field
    @property
    def zabbix_agent(self) -> str:
        """LEGACY: Serializes the zabbix agent status as a formatted string
        in legacy mode, and as-is in new mode.
        """
        # NOTE: use `self.active_available` instead of `self.zabbix_agent`
        if self.legacy_json_format:
            return self.get_active_status(with_code=True)
        return self.get_active_status()

    @field_serializer("status", when_used="json")
    def _LEGACY_status_serializer(
        self, v: Optional[str], _info: FieldSerializationInfo
    ) -> Optional[str]:
        """Serializes the monitoring status as a formatted string
        in legacy mode, and as-is in new mode.
        """
        if self.legacy_json_format:
            return get_monitoring_status(v, with_code=True)
        return v

    @field_validator("host", mode="before")  # TODO: add test for this
    @classmethod
    def _use_id_if_empty(cls, v: str, info: ValidationInfo) -> str:
        """In case the Zabbix API returns no host name, use the ID instead."""
        if not v:
            return f"Unknown (ID: {info.data.get('hostid')})"
        return v

    @field_validator("proxyid", mode="after")  # TODO: add test for this
    @classmethod
    def _proxyid_0_is_none(cls, v: str, info: ValidationInfo) -> Optional[str]:
        """Zabbix API can return 0 if host has no proxy.

        Convert to None, so we know the proxyid can always be used to
        look up the proxy, as well as in boolean contexts.
        """
        if not v or v == "0":
            return None
        return v

    def __cols_rows__(self) -> ColsRowsType:
        cols = [
            "HostID",
            "Name",
            "Host groups",
            "Templates",
            "Agent",
            "Maintenance",
            "Status",
            "Proxy",
        ]
        rows: RowsType = [
            [
                self.hostid,
                self.host,
                "\n".join([group.name for group in self.groups]),
                "\n".join([template.host for template in self.templates]),
                self.zabbix_agent,
                MaintenanceStatus.string_from_value(self.maintenance_status),
                MonitoringStatus.string_from_value(self.status),
                self.proxy.name if self.proxy else "",
            ]
        ]
        return cols, rows


class HostInterface(ZabbixAPIBaseModel):
    type: int
    ip: str
    dns: Optional[str] = None
    port: str
    useip: int
    main: int
    # Values not required for creation:
    interfaceid: Optional[str] = None
    available: Optional[int] = None
    hostid: Optional[str] = None
    bulk: Optional[int] = None

    @computed_field
    @property
    def connection_mode(self) -> str:
        """Returns the connection mode as a formatted string."""
        return InterfaceConnectionMode.string_from_value(self.useip)

    @computed_field
    @property
    def type_str(self) -> str:
        """Returns the interface type as a formatted string."""
        return InterfaceType.string_from_value(self.type)

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["ID", "Type", "IP", "DNS", "Port", "Mode", "Default", "Available"]
        rows: RowsType = [
            [
                self.interfaceid or "",
                str(InterfaceType(self.type).value),
                self.ip,
                self.dns or "",
                self.port,
                self.connection_mode,
                str(bool(self.main)),
                ActiveInterface.string_from_value(self.available),
            ]
        ]
        return cols, rows


class CreateHostInterfaceDetails(ZabbixAPIBaseModel):
    """Parameters for creating a host interface."""

    version: int
    bulk: Optional[int] = None
    community: Optional[str] = None
    max_repetitions: Optional[int] = None
    securityname: Optional[str] = None
    securitylevel: Optional[int] = None
    authpassphrase: Optional[str] = None
    privpassphrase: Optional[str] = None
    authprotocol: Optional[int] = None
    privprotocol: Optional[int] = None
    contextname: Optional[str] = None


class UpdateHostInterfaceDetails(ZabbixAPIBaseModel):
    """Parameters for updating a host interface."""

    version: Optional[int] = None
    bulk: Optional[int] = None
    community: Optional[str] = None
    max_repetitions: Optional[int] = None
    securityname: Optional[str] = None
    securitylevel: Optional[int] = None
    authpassphrase: Optional[str] = None
    privpassphrase: Optional[str] = None
    authprotocol: Optional[int] = None
    privprotocol: Optional[int] = None
    contextname: Optional[str] = None


class Proxy(ZabbixAPIBaseModel):
    """Zabbix Proxy object."""

    proxyid: str
    name: str = Field(..., validation_alias=AliasChoices("host", "name"))
    hosts: HostList = Field(default_factory=list)
    status: Optional[int] = None
    operating_mode: Optional[int] = None
    address: str = Field(
        validation_alias=AliasChoices(
            "address",  # >=7.0.0
            "proxy_address",  # <7.0.0
        )
    )
    proxy_groupid: Optional[str] = None  # >= 7.0
    compatibility: Optional[int] = None  # >= 7.0
    version: Optional[int] = None  # >= 7.0
    local_address: Optional[str] = None  # >= 7.0
    local_port: Optional[str] = None  # >= 7.0

    def __hash__(self) -> str:
        return self.proxyid  # kinda hacky, but lets us use it in dicts

    @computed_field
    @property
    def mode(self) -> str:
        """Returns the proxy mode as a formatted string."""
        if self.zabbix_version.release >= (7, 0, 0):
            cls = ProxyMode
        else:
            cls = ProxyModePre70
        return cls.string_from_value(self.operating_mode)

    @computed_field
    @property
    def compatibility_str(self) -> str:
        """Returns the proxy compatibility as a formatted string."""
        return ProxyCompatibility.string_from_value(self.compatibility)

    @property
    def compatibility_rich(self) -> str:
        """Returns the proxy compatibility as a Rich markup formatted string."""
        compat = self.compatibility_str
        if compat == "Current":
            style = Color.SUCCESS
        elif compat == "Outdated":
            style = Color.WARNING
        elif compat == "Unsupported":
            style = Color.ERROR
        else:
            style = Color.INFO
        return style(compat)


class ProxyGroup(ZabbixAPIBaseModel):
    """Zabbix Proxy Group object."""

    proxy_groupid: str
    name: str
    description: str
    failover_delay: str
    min_online: str  # 1-1000, may be a macro
    state: ProxyGroupState
    proxies: list[Proxy] = Field(default_factory=list)

    def __cols_rows__(self) -> ColsRowsType:
        cols = [
            "ID",
            "Name",
            "Description",
            "Failover Delay",
            "Minimum Available",
            "State",
            "Proxies",
        ]
        rows: RowsType = [
            [
                self.proxy_groupid,
                self.name,
                self.description,
                self.failover_delay,
                str(self.min_online),
                str(self.state),  # TODO: make prettier
                "\n".join(proxy.name for proxy in self.proxies),
            ]
        ]
        return cols, rows


class MacroBase(ZabbixAPIBaseModel):
    macro: str
    value: Optional[str] = None  # Optional in case secret value
    type: int
    """Macro type. 0 - text, 1 - secret, 2 - vault secret (>=7.0)"""
    description: str

    @computed_field
    @property
    def type_str(self) -> str:
        """Returns the macro type as a formatted string."""
        return MacroType.string_from_value(self.type)


class Macro(MacroBase):
    """Macro object. Known as 'host macro' in the Zabbix API."""

    hostid: str
    hostmacroid: str
    automatic: Optional[int] = None  # >= 7.0 only. 0 = user, 1 = discovery rule
    hosts: HostList = Field(default_factory=list)
    templates: list[Template] = Field(default_factory=list)


class GlobalMacro(MacroBase):
    """Global macro object."""

    globalmacroid: str


class Item(ZabbixAPIBaseModel):
    """Zabbix Item."""

    itemid: str
    delay: Optional[str] = None
    hostid: Optional[str] = None
    interfaceid: Optional[str] = None
    key: Optional[str] = Field(
        default=None, validation_alias=AliasChoices("key_", "key")
    )
    name: Optional[str] = None
    type: Optional[int] = None
    url: Optional[str] = None
    value_type: Optional[int] = None
    description: Optional[str] = None
    history: Optional[str] = None
    lastvalue: Optional[str] = None
    hosts: HostList = []

    @computed_field
    @property
    def type_str(self) -> str:
        """Returns the item type as a formatted string."""
        return ItemType.string_from_value(self.type)

    @computed_field
    @property
    def value_type_str(self) -> str:
        """Returns the item type as a formatted string."""
        return ValueType.string_from_value(self.value_type)

    @field_serializer("type")
    def _LEGACY_type_serializer(
        self, v: Optional[int], _info: FieldSerializationInfo
    ) -> Union[str, int, None]:
        """Serializes the item type as a formatted string in legacy JSON mode."""
        if self.legacy_json_format:
            return self.type_str
        return v

    @field_serializer("value_type")
    def _LEGACY_value_type_serializer(
        self, v: Optional[int], _info: FieldSerializationInfo
    ) -> Union[str, int, None]:
        """Serializes the item type as a formatted string in legacy JSON mode."""
        if self.legacy_json_format:
            return self.type_str
        return v

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["ID", "Name", "Key", "Type", "Interval", "History", "Description"]
        rows: RowsType = [
            [
                self.itemid,
                str(self.name),
                str(self.key),
                str(self.type_str),
                str(self.delay),
                str(self.history),
                str(self.description),
            ]
        ]
        return cols, rows


class Role(ZabbixAPIBaseModel):
    roleid: str
    name: str
    type: int
    readonly: int  # 0 = read-write, 1 = read-only


class MediaType(ZabbixAPIBaseModel):
    mediatypeid: str
    name: str
    type: int
    description: Optional[str] = None

    @computed_field
    @property
    def type_str(self) -> str:
        return MediaTypeType.string_from_value(self.type)

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["ID", "Name", "Type", "Description"]
        rows: RowsType = [
            [
                self.mediatypeid,
                self.name,
                self.type_str,
                self.description or "",
            ]
        ]
        return cols, rows


class UserMedia(ZabbixAPIBaseModel):
    """Media attached to a user object."""

    # https://www.zabbix.com/documentation/current/en/manual/api/reference/user/object#media
    mediatypeid: str
    sendto: str
    active: int = 0  # 0 = enabled, 1 = disabled (YES REALLY!)
    severity: int = 63  # all (1111 in binary - all bits set)
    period: str = "1-7,00:00-24:00"  # 24/7


class TimePeriod(ZabbixAPIBaseModel):
    period: int
    timeperiod_type: int
    start_date: Optional[datetime] = None
    start_time: Optional[int] = None
    every: Optional[int] = None
    dayofweek: Optional[int] = None
    day: Optional[int] = None
    month: Optional[int] = None

    @computed_field
    @property
    def timeperiod_type_str(self) -> str:
        """Returns the time period type as a formatted string."""
        return MaintenancePeriodType.string_from_value(self.timeperiod_type)

    @computed_field
    @property
    def period_str(self) -> str:
        return str(timedelta(seconds=self.period))

    @computed_field
    @property
    def start_time_str(self) -> str:
        return str(timedelta(seconds=self.start_time or 0))

    @computed_field
    @property
    def start_date_str(self) -> str:
        if self.start_date and self.start_date.year > 1970:  # hack to avoid 1970-01-01
            return self.start_date.strftime("%Y-%m-%d %H:%M")
        return ""

    @computed_field
    @property
    def month_str(self) -> list[str]:
        return get_maintenance_active_months(self.month)

    @computed_field
    @property
    def dayofweek_str(self) -> list[str]:
        return get_maintenance_active_days(self.dayofweek)

    @computed_field
    @property
    def every_str(self) -> str:
        return MaintenanceWeekType.string_from_value(self.every)

    def __cols_rows__(self) -> ColsRowsType:
        """Renders the table based on the time period type.

        Fields are added/removed based on the time period type.
        """
        # TODO: use MaintenancePeriodType enum for this
        if self.timeperiod_type == 0:
            return self._get_cols_rows_one_time()
        # TODO: add __cols_rows__ method for each timeperiod type
        # other timeperiod types here...
        else:
            return self._get_cols_rows_default()

    def _get_cols_rows_default(self) -> ColsRowsType:
        """Fallback for when we don't know the time period type."""
        cols = [
            "Type",
            "Duration",
            "Start date",
            "Start time",
            "Every",
            "Day of week",
            "Day",
            "Months",
        ]
        rows: RowsType = [
            [
                self.timeperiod_type_str,
                self.period_str,
                self.start_date_str,
                self.start_time_str,
                self.every_str,
                "\n".join(self.dayofweek_str),
                str(self.day),
                "\n".join(self.month_str),
            ]
        ]
        return cols, rows

    def _get_cols_rows_one_time(self) -> ColsRowsType:
        """Get the cols and rows for a one time schedule."""
        cols = ["Type", "Duration", "Start date"]
        rows: RowsType = [
            [
                self.timeperiod_type_str,
                self.period_str,
                self.start_date_str,
            ]
        ]
        return cols, rows


class ProblemTag(ZabbixAPIBaseModel):
    tag: str
    operator: Optional[int]
    value: Optional[str]


class Maintenance(ZabbixAPIBaseModel):
    maintenanceid: str
    name: str
    active_since: Optional[datetime] = None
    active_till: Optional[datetime] = None
    description: Optional[str] = None
    maintenance_type: Optional[int] = None
    tags_evaltype: Optional[int] = None
    timeperiods: list[TimePeriod] = []
    tags: list[ProblemTag] = []
    hosts: HostList = []
    hostgroups: list[HostGroup] = Field(
        default_factory=list, validation_alias=AliasChoices("groups", "hostgroups")
    )

    @field_validator("timeperiods", mode="after")
    @classmethod
    def _sort_time_periods(cls, v: list[TimePeriod]) -> list[TimePeriod]:
        """Cheeky hack to consistently render mixed time period types.

        This validator ensures time periods are sorted by complexity, so that the
        most complex ones are rendered first, thus adding all the necessary
        columns to the table when multiple time periods are rendered in aggregate.

        See: TimePeriod.__cols_rows__
        See: AggregateResult.__cols_rows__

        """
        # 0 = one time. We want those last.
        return sorted(v, key=lambda tp: tp.timeperiod_type, reverse=True)


class Event(ZabbixAPIBaseModel):
    eventid: str
    source: int
    object: int
    objectid: str
    acknowledged: int
    clock: Optional[datetime] = None
    name: str
    value: Optional[int] = None  # docs seem to imply this is optional
    severity: int
    # NYI:
    # r_eventid
    # c_eventid
    # cause_eventid
    # correlationid
    # userid
    # suppressed
    # opdata
    # urls

    @computed_field
    @property
    def age(self) -> Optional[str]:
        """Returns the age of the event as a formatted string."""
        return age_from_datetime(self.clock)

    @computed_field
    @property
    def status_str(self) -> str:
        return EventStatus.string_from_value(self.value)

    @computed_field
    @property
    def acknowledged_str(self) -> str:
        return AckStatus.string_from_value(self.acknowledged)

    @property
    def status_str_cell(self) -> str:
        """Formatted and styled status string for use in a table cell."""
        if self.status_str == "OK":
            color = "green"
        else:
            color = "red"
        return f"[{color}]{self.status_str}[/]"

    @property
    def acknowledged_str_cell(self) -> str:
        """Formatted and styled state string for use in a table cell."""
        if self.acknowledged_str == "Yes":
            color = "green"
        else:
            color = "red"
        return f"[{color}]{self.acknowledged_str}[/]"

    @field_serializer("value")
    def _LEGACY_value_serializer(
        self, v: Optional[int], _info: FieldSerializationInfo
    ) -> Union[str, int, None]:
        """Serializes the value field as a formatted string in legacy JSON mode"""
        if self.legacy_json_format:
            return self.status_str
        return v

    @field_serializer("acknowledged")
    def _LEGACY_acknowledged_serializer(
        self, v: int, _info: FieldSerializationInfo
    ) -> Union[str, int]:
        """Serializes the acknowledged field as a formatted string in legacy JSON mode"""
        if self.legacy_json_format:
            return self.acknowledged_str
        return v

    def __cols_rows__(self) -> ColsRowsType:
        cols = [
            "Event ID",
            "Trigger ID",
            "Name",
            "Last change",
            "Age",
            "Acknowledged",
            "Status",
        ]
        rows: RowsType = [
            [
                self.eventid,
                self.objectid,
                self.name,
                format_datetime(self.clock),
                self.age or "",
                self.acknowledged_str_cell,
                self.status_str_cell,
            ]
        ]
        return cols, rows


class Trigger(ZabbixAPIBaseModel):
    triggerid: str
    """Required for update operations."""
    description: Optional[str] = None
    """Required for create operations."""
    expression: Optional[str] = None
    """Required for create operations."""
    event_name: Optional[str] = None
    opdata: Optional[str] = None
    comments: Optional[str] = None
    error: Optional[str] = None
    flags: Optional[int] = None
    lastchange: Optional[datetime] = None
    priority: Optional[int] = None
    state: Optional[int] = None
    templateid: Optional[str] = None
    type: Optional[int] = None
    url: Optional[str] = None
    url_name: Optional[str] = None  # >6.0
    value: Optional[int] = None
    recovery_mode: Optional[int] = None
    recovery_expression: Optional[str] = None
    correlation_mode: Optional[int] = None
    correlation_tag: Optional[str] = None
    manual_close: Optional[int] = None
    uuid: Optional[str] = None
    hosts: list[Host] = []
    # NYI:
    # groups: List[HostGroup] = Field(
    #     default_factory=list, validation_alias=AliasChoices("groups", "hostgroups")
    # )
    # items
    # functions
    # dependencies
    # discoveryRule
    # lastEvent

    @computed_field
    @property
    def age(self) -> Optional[str]:
        """Returns the age of the event as a formatted string."""
        return age_from_datetime(self.lastchange)

    @computed_field
    @property
    def hostname(self) -> Optional[str]:
        """Returns the hostname of the trigger."""
        if self.hosts:
            return self.hosts[0].host
        return None

    @computed_field
    @property
    def severity(self) -> str:
        return TriggerPriority.string_from_value(self.priority)

    def __cols_rows__(self) -> ColsRowsType:
        cols = [
            "Trigger ID",
            "Host",
            "Description",
            "Severity",
            "Last Change",
            "Age",
        ]
        rows: RowsType = [
            [
                self.triggerid,
                self.hostname or "",
                self.description or "",
                self.severity,
                format_datetime(self.lastchange),
                self.age or "",
            ]
        ]
        return cols, rows


class Image(ZabbixAPIBaseModel):
    imageid: str
    name: str
    imagetype: int
    # NOTE: Optional so we can fetch an image without its data
    # This lets us get the IDs of all images without keeping the data in memory
    image: Optional[str] = None


class Map(ZabbixAPIBaseModel):
    sysmapid: str
    name: str
    height: int
    width: int
    backgroundid: Optional[str] = None  # will this be an empty string instead?
    # Other fields are omitted. We only use this for export and import.


class ImportRule(BaseModel):  # does not need to inherit from ZabbixAPIBaseModel
    createMissing: bool
    updateExisting: Optional[bool] = None
    deleteMissing: Optional[bool] = None


class ImportRules(ZabbixAPIBaseModel):
    discoveryRules: ImportRule
    graphs: ImportRule
    groups: Optional[ImportRule] = None  # < 6.2
    host_groups: Optional[ImportRule] = None  # >= 6.2
    hosts: ImportRule
    httptests: ImportRule
    images: ImportRule
    items: ImportRule
    maps: ImportRule
    mediaTypes: ImportRule
    template_groups: Optional[ImportRule] = None  # >= 6.2
    templateLinkage: ImportRule
    templates: ImportRule
    templateDashboards: ImportRule
    triggers: ImportRule
    valueMaps: ImportRule
    templateScreens: Optional[ImportRule] = None  # < 6.0
    applications: Optional[ImportRule] = None  # < 6.0
    screens: Optional[ImportRule] = None  # < 6.0

    model_config = ConfigDict(validate_assignment=True)

    @classmethod
    def get(
        cls,
        *,
        create_missing: bool = False,
        update_existing: bool = False,
        delete_missing: bool = False,
    ) -> ImportRules:
        """Create import rules given directives and Zabbix API version."""
        # Create/delete missing
        cd = ImportRule(createMissing=create_missing, deleteMissing=delete_missing)
        # Create/update missing
        cu = ImportRule(createMissing=create_missing, updateExisting=update_existing)
        # Create/update/delete missing
        cud = ImportRule(
            createMissing=create_missing,
            updateExisting=update_existing,
            deleteMissing=delete_missing,
        )
        rules = ImportRules(
            discoveryRules=cud,
            graphs=cud,
            hosts=cu,
            httptests=cud,
            images=cu,
            items=cud,
            maps=cu,
            mediaTypes=cu,
            templateLinkage=cd,
            templates=cu,
            templateDashboards=cud,
            triggers=cud,
            valueMaps=cud,
        )
        if cls.zabbix_version.release >= (6, 2, 0):
            rules.host_groups = cu
            rules.template_groups = cu
        else:
            rules.groups = ImportRule(createMissing=create_missing)

        if cls.zabbix_version.major < 6:
            rules.applications = cd
            rules.screens = cu
            rules.templateScreens = cud

        return rules


def resolve_forward_refs() -> None:
    """Certain models have forward references that need to be resolved.

    I.e. HostGroup has a field `hosts` that references the Host model,
    which is defined _after_ the HostGroup model. This function resolves
    those forward references so that we can serialize them properly.

    We do the simplest possible resolution here, which is to just
    rebuild all the models in the module. This is inefficient, but
    guarantees we won't have any runtime errors due to unresolved
    forward references.
    """
    for obj in globals().values():
        if not isinstance(obj, type):
            continue
        try:
            if not issubclass(obj, ZabbixAPIBaseModel):
                continue
        except TypeError:
            continue
        obj.model_rebuild()


resolve_forward_refs()



"""User result types."""

# NOTE: The user module was one of the first to be written, and thus
# most of the result rendering was implemented in the Pyzabbix models
# themselves instead of as result classes. That is why this module is
# (mostly) empty.
from __future__ import annotations

from zabbix_cli.models import AggregateResult
from zabbix_cli.models import ColsRowsType
from zabbix_cli.models import ColsType
from zabbix_cli.models import RowsType
from zabbix_cli.models import TableRenderable
from zabbix_cli.pyzabbix.types import Usergroup
from zabbix_cli.pyzabbix.types import UserMedia


class CreateNotificationUserResult(TableRenderable):
    """Result type for creating a notification user."""

    username: str
    userid: str
    media: list[UserMedia]
    usergroups: list[Usergroup]

    def __cols_rows__(self) -> ColsRowsType:
        cols: ColsType = [
            "User ID",
            "Username",
            "Media",
            "Usergroups",
        ]
        rows: RowsType = [
            [
                self.userid,
                self.username,
                AggregateResult(result=self.media).as_table(),
                "\n".join([ug.name for ug in self.usergroups]),
            ]
        ]
        return cols, rows



from __future__ import annotations

import pytest
from zabbix_cli.pyzabbix.enums import AckStatus
from zabbix_cli.pyzabbix.enums import ActiveInterface
from zabbix_cli.pyzabbix.enums import APIStr
from zabbix_cli.pyzabbix.enums import APIStrEnum
from zabbix_cli.pyzabbix.enums import DataCollectionMode
from zabbix_cli.pyzabbix.enums import EventStatus
from zabbix_cli.pyzabbix.enums import ExportFormat
from zabbix_cli.pyzabbix.enums import GUIAccess
from zabbix_cli.pyzabbix.enums import HostgroupFlag
from zabbix_cli.pyzabbix.enums import HostgroupType
from zabbix_cli.pyzabbix.enums import InterfaceConnectionMode
from zabbix_cli.pyzabbix.enums import InterfaceType
from zabbix_cli.pyzabbix.enums import InventoryMode
from zabbix_cli.pyzabbix.enums import ItemType
from zabbix_cli.pyzabbix.enums import MacroType
from zabbix_cli.pyzabbix.enums import MaintenanceStatus
from zabbix_cli.pyzabbix.enums import MaintenanceType
from zabbix_cli.pyzabbix.enums import MaintenanceWeekType
from zabbix_cli.pyzabbix.enums import MonitoredBy
from zabbix_cli.pyzabbix.enums import MonitoringStatus
from zabbix_cli.pyzabbix.enums import ProxyCompatibility
from zabbix_cli.pyzabbix.enums import ProxyGroupState
from zabbix_cli.pyzabbix.enums import ProxyMode
from zabbix_cli.pyzabbix.enums import ProxyModePre70
from zabbix_cli.pyzabbix.enums import SNMPAuthProtocol
from zabbix_cli.pyzabbix.enums import SNMPPrivProtocol
from zabbix_cli.pyzabbix.enums import SNMPSecurityLevel
from zabbix_cli.pyzabbix.enums import TriggerPriority
from zabbix_cli.pyzabbix.enums import UsergroupPermission
from zabbix_cli.pyzabbix.enums import UserRole
from zabbix_cli.pyzabbix.enums import ValueType

APISTR_ENUMS = [
    AckStatus,
    ActiveInterface,
    DataCollectionMode,
    EventStatus,
    GUIAccess,
    HostgroupFlag,
    HostgroupType,
    InterfaceConnectionMode,
    InterfaceType,
    InventoryMode,
    ItemType,
    MacroType,
    MaintenanceStatus,
    MaintenanceType,
    MaintenanceWeekType,
    MonitoredBy,
    MonitoringStatus,
    ProxyCompatibility,
    ProxyGroupState,
    ProxyMode,
    ProxyModePre70,
    SNMPSecurityLevel,
    SNMPAuthProtocol,
    SNMPPrivProtocol,
    TriggerPriority,
    UsergroupPermission,
    UserRole,
    ValueType,
]


@pytest.mark.parametrize("enum", APISTR_ENUMS)
def test_apistrenum(enum: type[APIStrEnum]) -> None:
    assert enum.__members__
    members = list(enum)
    assert members
    for member in members:
        # Narrow down type
        assert isinstance(member, enum)
        assert isinstance(member.value, str)
        assert isinstance(member.value, APIStr)

        # Methods
        assert member.as_api_value() is not None
        assert member.__choice_name__ is not None
        assert member.__fmt_name__()  # non-empty string

        # Test instantiation
        assert enum(member) == member
        assert enum(member.value) == member
        # NOTE: to support multiple versions of the Zabbix API, some enums
        # have multiple members with the same API value, and we cannot blindly
        # test instantiation with the API value for those specific enums.
        # To not overcomplicate things, we just skip that test for the affected members
        if member in (SNMPPrivProtocol.AES, SNMPPrivProtocol.AES128):
            continue
        assert enum(member.as_api_value()) == member
        assert enum(member.value.api_value) == member

        # Test string_from_value
        for value in [member.as_api_value(), member.value]:
            s = enum.string_from_value(value)
            if member.name != "UNKNOWN":
                assert "Unknown" not in s, f"{value} can't be converted to string"
            assert s == member.as_status()


def test_interfacetype() -> None:
    # We already test normal behavior in test_apistrenum, check special behavior here
    for member in InterfaceType:
        assert member.value.metadata
        assert member.get_port()
    assert InterfaceType.AGENT.get_port() == "10050"
    assert InterfaceType.SNMP.get_port() == "161"
    assert InterfaceType.IPMI.get_port() == "623"
    assert InterfaceType.JMX.get_port() == "12345"


def test_exportformat() -> None:
    assert ExportFormat.PHP not in ExportFormat.get_importables()
    assert ExportFormat("json") == ExportFormat("JSON")
    assert ExportFormat("xml") == ExportFormat("XML")
    assert ExportFormat("yaml") == ExportFormat("YAML")
    assert ExportFormat.JSON in ExportFormat.get_importables()
    assert ExportFormat.XML in ExportFormat.get_importables()
    assert ExportFormat.YAML in ExportFormat.get_importables()



from __future__ import annotations

from zabbix_cli.output.style import Color


def test_color_call() -> None:
    assert Color.INFO("test") == "[default]test[/]"
    assert Color.SUCCESS("test") == "[green]test[/]"
    assert Color.WARNING("test") == "[yellow]test[/]"
    assert Color.ERROR("test") == "[red]test[/]"

    # Ensure it doesnt break normal instantiation with a value
    assert Color("default") == Color.INFO
    assert Color("yellow") == Color.WARNING



"""Defines directories for the application.

Follows the XDG Base Directory Specification on Linux: <https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html>
See <https://pypi.org/project/platformdirs/> for other platforms.
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import NamedTuple

from platformdirs import PlatformDirs

from zabbix_cli.__about__ import APP_NAME
from zabbix_cli.__about__ import AUTHOR

logger = logging.getLogger(__name__)


_PLATFORM_DIR = PlatformDirs(APP_NAME, AUTHOR)

CONFIG_DIR = _PLATFORM_DIR.user_config_path
"""Directory for user configuration files."""

DATA_DIR = _PLATFORM_DIR.user_data_path
"""Directory for user data files."""

LOGS_DIR = _PLATFORM_DIR.user_log_path
"""Directory to store user log files."""

SITE_CONFIG_DIR = _PLATFORM_DIR.site_config_path
"""Directory for site-wide configuration files, i.e. `/etc/xdg/zabbix-cli`."""

EXPORT_DIR = DATA_DIR / "exports"
"""Directory to store exported data."""


class Directory(NamedTuple):
    name: str
    path: Path
    required: bool = False
    create: bool = True
    """Do not log/print on failure to create directory."""


DIRS = [
    Directory("Config", CONFIG_DIR),
    Directory("Data", DATA_DIR),
    Directory("Logs", LOGS_DIR),
    # Don't create site config directory by default (root-only)
    Directory("Site Config", SITE_CONFIG_DIR, create=False),
    # Exports directory is created on demand
    Directory("Exports", EXPORT_DIR, create=False),
]


def init_directories() -> None:
    """Create required directories for the application to function."""
    from .output.console import error
    from .output.console import exit_err

    for directory in DIRS:
        if directory.path.exists() or not directory.create:
            logger.debug(
                "Skipping creating directory '%s'. Exists: %s",
                directory.path,
                directory.path.exists(),
            )
            continue
        try:
            directory.path.mkdir(parents=True)
        except Exception as e:
            message = (
                f"Failed to create {directory.name} directory {directory.path}: {e}"
            )
            if directory.required:
                exit_err(message, exc_info=True)
            else:
                error(message, exc_info=True)



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING
from typing import Optional
from typing import Union

import typer

from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.commands.common.args import ARG_GROUP_NAMES_OR_IDS
from zabbix_cli.commands.common.args import ARG_TEMPLATE_NAMES_OR_IDS
from zabbix_cli.output.console import error
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import info
from zabbix_cli.output.console import success
from zabbix_cli.output.formatting.grammar import pluralize as p
from zabbix_cli.output.render import render_result
from zabbix_cli.pyzabbix.enums import UsergroupPermission
from zabbix_cli.utils.args import parse_hostgroups_arg
from zabbix_cli.utils.args import parse_list_arg
from zabbix_cli.utils.args import parse_templategroups_arg
from zabbix_cli.utils.args import parse_templates_arg

if TYPE_CHECKING:
    from zabbix_cli.pyzabbix.types import HostGroup
    from zabbix_cli.pyzabbix.types import TemplateGroup


HELP_PANEL = "Template Group"


# Changed in V3: Changed name to reflect introduction of template groups in >=6.2
@app.command("add_template_to_group", rich_help_panel=HELP_PANEL)
@app.command(
    # old name for backwards compatibility
    "link_template_to_hostgroup",
    hidden=True,
    deprecated=True,
    rich_help_panel=HELP_PANEL,
    help="DEPRECATED: Use add_template_to_group instead.",
)
def add_template_to_group(
    ctx: typer.Context,
    template_names_or_ids: str = ARG_TEMPLATE_NAMES_OR_IDS,
    group_names_or_ids: str = ARG_GROUP_NAMES_OR_IDS,
    strict: bool = typer.Option(
        False,
        "--strict",
        help="Fail if any host groups or templates aren't found. Should not be used in conjunction with wildcards.",
    ),
) -> None:
    """Add templates to groups.

    [bold]NOTE:[/] Group arguments are interpreted as template groups in >= 6.2,
    otherwise as host groups.
    """
    from zabbix_cli.commands.results.template import TemplateGroupResult

    groups: Union[list[HostGroup], list[TemplateGroup]]
    if app.state.client.version.release >= (6, 2, 0):
        groups = parse_templategroups_arg(app, group_names_or_ids, strict=strict)
    else:
        groups = parse_hostgroups_arg(app, group_names_or_ids, strict=strict)

    templates = parse_templates_arg(app, template_names_or_ids, strict=strict)

    with app.state.console.status("Adding templates..."):
        app.state.client.link_templates_to_groups(templates, groups)
    render_result(TemplateGroupResult.from_result(templates, groups))
    success(f"Added {len(templates)} templates to {len(groups)} groups.")


@app.command(
    "create_templategroup",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Create a template group with default user group permissions",
            "create_templategroup 'My Template Group'",
        ),
        Example(
            "Create a template group with specific RO and RW groups",
            "create_templategroup 'My Template Group' --ro-groups users --rw-groups admins",
        ),
        Example(
            "Create a template group with no user group permissions",
            "create_templategroup 'My Template Group' --no-usergroup-permissions",
        ),
    ],
)
def create_templategroup(
    ctx: typer.Context,
    templategroup: str = typer.Argument(
        help="Name of the group.",
        show_default=False,
    ),
    rw_groups: Optional[str] = typer.Option(
        None,
        help="User group(s) to give read/write permissions. Comma-separated.",
    ),
    ro_groups: Optional[str] = typer.Option(
        None,
        help="User group(s) to give read-only permissions. Comma-separated.",
    ),
    use_default_usergroups: bool = typer.Option(
        True,
        "--default-usergroups/--no-default-usergroups",
        "--usergroup-permissions/--no-usergroup-permissions",
        help="Assign default user group permissions from configs.",
    ),
) -> None:
    """Create a new template group.

    Assigns default user group permissions by default.

    * [option]--rw-groups[/] defaults to config option [configopt]app.default_admin_usergroups[/].
    * [option]--ro-groups[/] defaults to config option [configopt]app.default_create_user_usergroups[/].
    * Use [option]--no-usergroup-permissions[/] to create a group without any user group permissions.

    [b]NOTE:[/] Calls [command]create_hostgroup[/] for Zabbix versions < 6.2.0.
    """
    from zabbix_cli.models import Result

    if app.state.client.version.release < (6, 2, 0):
        from zabbix_cli.commands.hostgroup import create_hostgroup

        create_hostgroup(
            hostgroup=templategroup,
            rw_groups=rw_groups,
            ro_groups=ro_groups,
            use_default_usergroups=use_default_usergroups,
        )
        return

    groupid = app.state.client.create_templategroup(templategroup)

    app_config = app.state.config.app

    rw_grps: list[str] = parse_list_arg(rw_groups)
    ro_grps: list[str] = parse_list_arg(ro_groups)
    if use_default_usergroups:
        rw_grps.extend(app_config.commands.create_templategroup.rw_groups)
        ro_grps.extend(app_config.commands.create_templategroup.ro_groups)

    try:
        # Admin group(s) gets Read/Write
        for usergroup in rw_grps:
            app.state.client.update_usergroup_rights(
                usergroup,
                [templategroup],
                UsergroupPermission.READ_WRITE,
                hostgroup=False,
            )
            info(f"Assigned Read/Write permission for user group {usergroup!r}")
        # Default group(s) gets Read
        for usergroup in ro_grps:
            app.state.client.update_usergroup_rights(
                usergroup,
                [templategroup],
                UsergroupPermission.READ_ONLY,
                hostgroup=False,
            )
            info(f"Assigned Read-only permission for user group {usergroup!r}")

    except Exception as e:
        error(f"Failed to assign permissions to template group {templategroup!r}: {e}")
        info("Deleting template group...")
        app.state.client.delete_templategroup(groupid)
        exit_err(f"Failed to create template group {templategroup!r}.")

    render_result(
        Result(message=f"Created template group {templategroup} ({groupid}).")
    )


@app.command("extend_templategroup", rich_help_panel=HELP_PANEL)
def extend_templategroup(
    ctx: typer.Context,
    src_group: str = typer.Argument(
        help="Group to get templates from.",
        show_default=False,
    ),
    dest_group: str = typer.Argument(
        help="Group(s) to add templates to. Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Show groups and templates without copying.",
    ),
) -> None:
    """Add all templates from a group to other groups.

    Interprets the source group as a template group in >= 6.2, otherwise as a host group.

    Does not modify the source group or its templates.
    To remove the templates from the source group, use the [command]move_templates[/] command instead.
    """
    from zabbix_cli.commands.results.templategroup import ExtendTemplateGroupResult

    dest_arg = parse_list_arg(dest_group)

    src: Union[HostGroup, TemplateGroup]
    dest: Union[list[HostGroup], list[TemplateGroup]]
    if app.state.client.version.release > (6, 2, 0):
        src = app.state.client.get_templategroup(src_group, select_templates=True)
        dest = app.state.client.get_templategroups(
            *dest_arg, select_templates=True, search=True
        )
    else:
        src = app.state.client.get_hostgroup(src_group, select_templates=True)
        dest = app.state.client.get_hostgroups(
            *dest_arg, select_templates=True, search=True
        )

    if not src.templates:
        exit_err(f"No templates found in {src_group!r}")
    if not dest:
        exit_err(f"No groups found matching {dest_group!r}")

    if not dryrun:
        app.state.client.add_templates_to_groups(src.templates, dest)
    else:
        info("Would copy the following templates:")
    render_result(ExtendTemplateGroupResult.from_result(src, dest, src.templates))
    if not dryrun:
        success(
            f"Copied {len(src.templates)} templates from {src.name} to {len(dest)} groups."
        )


@app.command("move_templates", rich_help_panel=HELP_PANEL)
def move_templates(
    src_group: str = typer.Argument(
        help="Group to move templates from.",
        show_default=False,
    ),
    dest_group: str = typer.Argument(
        help="Group to move templates to.",
        show_default=False,
    ),
    rollback: bool = typer.Option(
        True,
        "--rollback/--no-rollback",
        help="Rollback changes if templates cannot be removed from source group afterwards.",
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Show templates and groups without making changes.",
    ),
) -> None:
    """Move all templates from one group to another."""
    from zabbix_cli.commands.results.templategroup import MoveTemplatesResult

    src: Union[HostGroup, TemplateGroup]
    dest: Union[HostGroup, TemplateGroup]

    if app.state.client.version.release < (6, 2, 0):
        src = app.state.client.get_hostgroup(src_group, select_templates=True)
        dest = app.state.client.get_hostgroup(dest_group, select_templates=True)
    else:
        src = app.state.client.get_templategroup(src_group, select_templates=True)
        dest = app.state.client.get_templategroup(dest_group, select_templates=True)

    if not src.templates:
        exit_err(f"No templates found in template group {src_group!r}.")

    if dryrun:
        info(f"Would copy {len(src.templates)} templates from {src.name!r}:")
    else:
        app.state.client.add_templates_to_groups(
            src.templates,
            # Clunky typing semantics here:
            # We cannot prove that list is homogenous ("list[HostGroup] | list[TemplateGroup]")
            # because inferred type is "list[HostGroup | TemplateGroup]""
            [dest],  # pyright: ignore[reportArgumentType]
        )
        info(f"Added templates to {dest.name!r}.")
        try:
            app.state.client.remove_templates_from_groups(
                src.templates,
                [src],  # pyright: ignore[reportArgumentType] # ditto
            )
        except Exception as e:
            if rollback:
                error(
                    f"Failed to remove hosts from {src.name!r}. Attempting to roll back changes."
                )
                app.state.client.remove_templates_from_groups(
                    src.templates,
                    [dest],  # pyright: ignore[reportArgumentType] # ditto
                )
            raise e
        else:
            info(f"Removed templates from {src.name!r}.")

    render_result(MoveTemplatesResult.from_result(src, dest))
    if not dryrun:
        success(
            f"Moved {len(src.templates)} templates from {src.name!r} to {dest.name!r}."
        )


@app.command("remove_templategroup", rich_help_panel=HELP_PANEL)
def remove_templategroup(
    ctx: typer.Context,
    templategroup: str = typer.Argument(
        help="Name of the group to delete.",
        show_default=False,
    ),
) -> None:
    """Delete a template group.

    NOTE: Calls [command]remove_hostgroup[/] for Zabbix <6.2.
    """
    from zabbix_cli.models import Result

    if app.state.client.version.release < (6, 2, 0):
        from zabbix_cli.commands.hostgroup import delete_hostgroup

        delete_hostgroup(hostgroup=templategroup)
        return

    app.state.client.delete_templategroup(templategroup)
    render_result(Result(message=f"Deleted template group {templategroup!r}."))


# Changed in V3: Changed name to reflect introduction of template groups in >=6.2
@app.command(
    "remove_template_from_group",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Remove one template from one group",
            "remove_template_from_group 'Apache by HTTP' foo_group",
        ),
        Example(
            "Remove many templates from many groups",
            "remove_template_from_group 'Apache by HTTP,HAProxy by Zabbix agent' foo_group,bar_group",
        ),
        Example(
            "Remove all templates starting with 'Apache' from a group",
            "remove_template_from_group 'Apache*' foo_group",
        ),
        Example(
            "Remove all templates containing 'HTTP' from all groups",
            "remove_template_from_group '*HTTP*' '*'",
        ),
    ],
)
@app.command(
    # old name for backwards compatibility
    "unlink_template_from_hostgroup",
    hidden=True,
    deprecated=True,
    rich_help_panel=HELP_PANEL,
    help="Use `remove_template_from_group` instead.",
)
def remove_template_from_group(
    ctx: typer.Context,
    template_names_or_ids: str = ARG_TEMPLATE_NAMES_OR_IDS,
    group_names_or_ids: str = ARG_GROUP_NAMES_OR_IDS,
    strict: bool = typer.Option(
        False,
        "--strict",
        help="Fail if any host groups or templates aren't found. Should not be used in conjunction with wildcards.",
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Preview changes.",
    ),
    # TODO: add toggle for NOT clearing when unlinking?
) -> None:
    """Remove templates from groups."""
    from zabbix_cli.commands.results.template import RemoveTemplateFromGroupResult
    from zabbix_cli.models import AggregateResult

    groups: Union[list[HostGroup], list[TemplateGroup]]
    if app.state.client.version.release >= (6, 2, 0):
        groups = parse_templategroups_arg(
            app, group_names_or_ids, strict=strict, select_templates=True
        )
    else:
        groups = parse_hostgroups_arg(
            app, group_names_or_ids, strict=strict, select_templates=True
        )

    templates = parse_templates_arg(app, template_names_or_ids, strict=strict)

    if not dryrun:
        with app.state.console.status("Removing templates from groups..."):
            # LEGACY: This used to also pass the templateids to templateids_clear,
            # which would unlink and clear all the templates from each other
            # This was a mistake, and has been removed in V3.
            # Users should use `unlink_templates_from_templates` for that.
            app.state.client.remove_templates_from_groups(
                templates,
                groups,
            )
    result: list[RemoveTemplateFromGroupResult] = []
    for group in groups:
        r = RemoveTemplateFromGroupResult.from_result(templates, group)
        if not r.templates:
            continue
        result.append(r)

    total_templates = len(set(chain.from_iterable((r.templates) for r in result)))
    total_groups = len(result)

    render_result(AggregateResult(result=result, empty_ok=True))
    base_msg = f"{p('template', total_templates)} from {p('group', total_groups)}"
    if dryrun:
        info(f"Would remove {base_msg}.")
    else:
        success(f"Removed {base_msg}.")


@app.command("show_templategroup", rich_help_panel=HELP_PANEL)
def show_templategroup(
    ctx: typer.Context,
    templategroup: str = typer.Argument(
        help="Name of the group to show. Supports wildcards.",
        show_default=False,
    ),
    templates: bool = typer.Option(
        True,
        "--templates/--no-templates",
        help="Show/hide templates associated with the group.",
    ),
) -> None:
    """Show details for a template group."""
    from zabbix_cli.commands.results.templategroup import ShowTemplateGroupResult

    tg: HostGroup | TemplateGroup
    if app.state.client.version.release < (6, 2, 0):
        tg = app.state.client.get_hostgroup(
            templategroup, search=True, select_templates=templates
        )
    else:
        tg = app.state.client.get_templategroup(
            templategroup, search=True, select_templates=templates
        )
    render_result(ShowTemplateGroupResult.from_result(tg, show_templates=templates))


@app.command(
    "show_templategroups",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Show all template groups",
            "show_templategroups",
        ),
        Example(
            "Show all template groups starting with 'Web-'",
            "show_templategroups 'Web-*'",
        ),
        Example(
            "Show template groups with 'web' in the name",
            "show_templategroups '*web*'",
        ),
    ],
)
def show_templategroups(
    ctx: typer.Context,
    name: Optional[str] = typer.Argument(
        None,
        help="Name of template group(s). Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    templates: bool = typer.Option(
        True,
        "--templates/--no-templates",
        help="Show/hide templates associated with each group.",
    ),
) -> None:
    """Show template groups.

    Fetches all groups by default, but can be filtered by name.
    """
    from zabbix_cli.commands.results.templategroup import ShowTemplateGroupResult
    from zabbix_cli.models import AggregateResult

    names = parse_list_arg(name)

    groups: Union[list[HostGroup], list[TemplateGroup]]
    with app.status("Fetching template groups..."):
        if app.state.client.version.release < (6, 2, 0):
            groups = app.state.client.get_hostgroups(
                *names,
                search=True,
                select_templates=True,
                sort_field="name",
                sort_order="ASC",
            )
        else:
            groups = app.state.client.get_templategroups(
                *names,
                search=True,
                select_templates=True,
                sort_field="name",
                sort_order="ASC",
            )

    # Sort by name before rendering
    groups = sorted(groups, key=lambda tg: tg.name)  # type: ignore # unable to infer that type doesn't change?

    render_result(
        AggregateResult(
            result=[ShowTemplateGroupResult.from_result(tg, templates) for tg in groups]
        )
    )


@app.command("show_templategroup_permissions", rich_help_panel=HELP_PANEL)
def show_templategroup_permissions(
    templategroups: str = typer.Argument(
        help="Template group name(s). Comma-separated. Supports wildcards.",
        show_default=False,
    ),
) -> None:
    """Show usergroups with permissions for the given templategroup.

    Shows permissions for all template groups by default.
    """
    from zabbix_cli.commands.results.templategroup import TemplateGroupPermissions
    from zabbix_cli.models import AggregateResult

    tg_names = parse_list_arg(templategroups)

    with app.status("Fetching template groups..."):
        usergroups = app.state.client.get_usergroups()
        tgs = app.state.client.get_templategroups(
            *tg_names,
            sort_field="name",
            sort_order="ASC",
            select_templates=False,
            search=True,
        )

    result: list[TemplateGroupPermissions] = []
    for tg in tgs:
        permissions: list[str] = []
        for usergroup in usergroups:
            if app.api_version >= (6, 2, 0):
                rights = usergroup.templategroup_rights
            else:
                rights = usergroup.rights
            for right in rights:
                if right.id == tg.groupid:
                    perm = UsergroupPermission.string_from_value(right.permission)
                    permissions.append(f"{usergroup.name} ({perm})")
                    break
        result.append(
            TemplateGroupPermissions(
                groupid=tg.groupid,
                name=tg.name,
                permissions=permissions,
            )
        )
    return render_result(AggregateResult(result=result))



from __future__ import annotations

import sys
from pathlib import Path
from typing import Any

sys.path.append(Path(__file__).parent.as_posix())

import docs.scripts.gen_commands as gen_commands  # noqa
import gen_cli_data  # noqa
import gen_cli_options  # noqa
import gen_command_list  # noqa
import gen_formats  # noqa
import gen_config_data  # noqa


def main(*args: Any, **kwargs: Any) -> None:
    for mod in [
        gen_cli_data,
        gen_cli_options,
        gen_command_list,
        gen_commands,
        gen_formats,
        gen_config_data,
    ]:
        mod.main()



from __future__ import annotations

import sys
from pathlib import Path

import yaml  # type: ignore
from zabbix_cli.config.constants import OutputFormat

sys.path.append(Path(__file__).parent.as_posix())

from common import DATA_DIR  # noqa


def main() -> None:
    fmts = [fmt.value for fmt in OutputFormat]

    with open(DATA_DIR / "formats.yaml", "w") as f:
        yaml.dump(fmts, f, default_flow_style=False)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING

from rich import box
from rich.table import Table

if TYPE_CHECKING:
    from zabbix_cli.models import ColsType
    from zabbix_cli.models import RowsType


def get_table(
    cols: ColsType,
    rows: RowsType,
    title: str | None = None,
    *,
    show_lines: bool = True,
    box: box.Box = box.ROUNDED,
) -> Table:
    """Returns a Rich table given a list of columns and rows."""
    table = Table(title=title, box=box, show_lines=show_lines)
    for col in cols:
        table.add_column(col, overflow="fold")
    for row in rows:
        # We might have subtables in the rows.
        # If they have no rows, we don't want to render them.
        row = [cell if not isinstance(cell, Table) or cell.rows else "" for cell in row]
        table.add_row(*row)
    return table



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING
from typing import Any
from typing import Union

import rich
import rich.box
from pydantic import Field
from pydantic import computed_field
from pydantic import field_validator
from pydantic import model_serializer

from zabbix_cli.models import MetaKey
from zabbix_cli.models import TableRenderable
from zabbix_cli.pyzabbix.enums import GUIAccess
from zabbix_cli.pyzabbix.enums import UsergroupPermission
from zabbix_cli.pyzabbix.enums import UsergroupStatus
from zabbix_cli.pyzabbix.types import HostGroup
from zabbix_cli.pyzabbix.types import TemplateGroup
from zabbix_cli.pyzabbix.types import Usergroup
from zabbix_cli.pyzabbix.types import ZabbixRight

if TYPE_CHECKING:
    from zabbix_cli.models import ColsRowsType
    from zabbix_cli.models import RowContent
    from zabbix_cli.models import RowsType


class UgroupUpdateUsersResult(TableRenderable):
    usergroups: list[str]
    users: list[str]

    def __cols_rows__(self) -> ColsRowsType:
        return (
            ["Usergroups", "Users"],
            [["\n".join(self.usergroups), ", ".join(self.users)]],
        )


class UsergroupAddUsers(UgroupUpdateUsersResult):
    __title__ = "Added Users"


class UsergroupRemoveUsers(UgroupUpdateUsersResult):
    __title__ = "Removed Users"


class AddUsergroupPermissionsResult(TableRenderable):
    usergroup: str
    hostgroups: list[str]
    templategroups: list[str]
    permission: UsergroupPermission

    @computed_field
    @property
    def permission_str(self) -> str:
        # FIXME: remove this? Serializing a Choice enum should dump the same value?
        return self.permission.as_status()

    def __cols_rows__(self) -> ColsRowsType:
        return (
            [
                "Usergroup",
                "Host Groups",
                "Template Groups",
                "Permission",
            ],
            [
                [
                    self.usergroup,
                    ", ".join(self.hostgroups),
                    ", ".join(self.templategroups),
                    self.permission_str,
                ],
            ],
        )


class ShowUsergroupResult(TableRenderable):
    usrgrpid: str = Field(..., json_schema_extra={MetaKey.HEADER: "ID"})
    name: str
    gui_access: str = Field(..., json_schema_extra={MetaKey.HEADER: "GUI Access"})
    status: str
    users: list[str] = Field(
        default_factory=list, json_schema_extra={MetaKey.JOIN_CHAR: ", "}
    )

    @classmethod
    def from_usergroup(cls, usergroup: Usergroup) -> ShowUsergroupResult:
        return cls(
            name=usergroup.name,
            usrgrpid=usergroup.usrgrpid,
            gui_access=usergroup.gui_access_str,
            status=usergroup.users_status_str,
            users=[user.username for user in usergroup.users],
        )

    @field_validator("gui_access")
    @classmethod
    def _validate_gui_access(cls, v: Any) -> str:
        if isinstance(v, int):
            return GUIAccess.string_from_value(v, with_code=False)
        return v

    @field_validator("status")
    @classmethod
    def _validate_status(cls, v: Any) -> str:
        if isinstance(v, int):
            return UsergroupStatus.string_from_value(v, with_code=False)
        return v


class GroupRights(TableRenderable):
    """Subtable for displaying group rights."""

    __box__ = rich.box.MINIMAL

    groups: Union[dict[str, HostGroup], dict[str, TemplateGroup]] = Field(
        default_factory=dict,
    )

    rights: list[ZabbixRight] = Field(
        default_factory=list,
        description="Group rights for the user group.",
    )

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["Name", "Permission"]
        rows: RowsType = []
        for right in self.rights:
            group = self.groups.get(right.id, None)
            if group:
                group_name = group.name
            else:
                group_name = "Unknown"
            rows.append([group_name, str(UsergroupPermission(right.permission))])
        return cols, rows


class ShowUsergroupPermissionsResult(TableRenderable):
    usrgrpid: str
    name: str
    hostgroups: dict[str, HostGroup] = Field(
        default_factory=dict,
        exclude=True,
        description="Host groups the user group has access to. Used to render host group rights.",
    )
    templategroups: dict[str, TemplateGroup] = Field(
        default_factory=dict,
        exclude=True,
        description="Mapping of all template groups. Used to render template group rights.",
    )
    hostgroup_rights: list[ZabbixRight] = []
    templategroup_rights: list[ZabbixRight] = []

    @model_serializer
    def model_ser(self) -> dict[str, Any]:
        """LEGACY: Include the permission strings in the serialized output if
        we have legacy JSON output enabled.
        """
        d: dict[str, Any] = {
            "usrgrpid": self.usrgrpid,
            "name": self.name,
            "hostgroup_rights": self.hostgroup_rights,
            "templategroup_rights": self.templategroup_rights,
        }
        if self.legacy_json_format:
            d["permissions"] = self.permissions
            d["usergroupid"] = self.usrgrpid
        return d

    @property
    def permissions(self) -> list[str]:
        """LEGACY: The field `hostgroup_rights` was called `permissions` in V2."""
        r: list[str] = []

        def permission_str(
            right: ZabbixRight, groups: Mapping[str, Union[HostGroup, TemplateGroup]]
        ) -> str:
            group = groups.get(right.id, None)
            if group:
                group_name = group.name
            else:
                group_name = "Unknown"
            perm = UsergroupPermission.string_from_value(
                right.permission, with_code=True
            )
            return f"{group_name} ({perm})"

        for right in self.hostgroup_rights:
            r.append(permission_str(right, self.hostgroups))
        for right in self.templategroup_rights:
            r.append(permission_str(right, self.templategroups))
        return r

    @classmethod
    def from_usergroup(
        cls,
        usergroup: Usergroup,
        hostgroups: list[HostGroup],
        templategroups: list[TemplateGroup],
    ) -> ShowUsergroupPermissionsResult:
        cls.model_rebuild()  # TODO: can we avoid this?
        res = cls(
            usrgrpid=usergroup.usrgrpid,
            name=usergroup.name,
            hostgroups={hg.groupid: hg for hg in hostgroups},
            templategroups={tg.groupid: tg for tg in templategroups},
            templategroup_rights=usergroup.templategroup_rights,
        )
        if res.zabbix_version.release >= (6, 2, 0):
            res.hostgroup_rights = usergroup.hostgroup_rights
        else:
            res.hostgroup_rights = usergroup.rights
        return res

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["ID", "Name", "Host Groups"]
        row: RowContent = [self.usrgrpid, self.name]

        # Host group rights table
        row.append(
            GroupRights(groups=self.hostgroups, rights=self.hostgroup_rights).as_table()
        )

        # Template group rights table
        if self.zabbix_version.release >= (6, 2, 0):
            cols.append("Template Groups")
            row.append(
                GroupRights(
                    groups=self.templategroups, rights=self.templategroup_rights
                ).as_table()
            )
        return cols, [row]


class RemoveUsergroupResult(TableRenderable):
    """Result type for removing user groups."""

    __title__ = "Removed User Groups"

    usergroups: list[Usergroup] = []

    def __cols_rows__(self) -> ColsRowsType:
        return (
            ["User group", "Users"],
            [
                [ug.name, ", ".join(u.username for u in ug.users)]
                for ug in self.usergroups
            ],
        )

    @classmethod
    def from_result(cls, usergroups: list[Usergroup]) -> RemoveUsergroupResult:
        return cls(usergroups=usergroups)



#!/usr/bin/env python
#
# Authors:
# rafael@e-mc2.net / https://e-mc2.net/
#
# Copyright (c) 2014-2024 USIT-University of Oslo
#
# This file is part of Zabbix-cli
# https://github.com/unioslo/zabbix-cli
#
# Zabbix-CLI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Zabbix-CLI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Zabbix-CLI.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Optional

import typer

from zabbix_cli.__about__ import __version__
from zabbix_cli.app import app
from zabbix_cli.config.constants import BulkRunnerMode
from zabbix_cli.config.constants import OutputFormat
from zabbix_cli.config.utils import get_config
from zabbix_cli.logs import configure_logging
from zabbix_cli.logs import logger
from zabbix_cli.state import get_state

if TYPE_CHECKING:
    from typing import Any


def run_repl(ctx: typer.Context) -> None:
    from rich.console import Group
    from rich.panel import Panel

    from zabbix_cli.output.console import console
    from zabbix_cli.output.style import green
    from zabbix_cli.repl.repl import repl as start_repl
    from zabbix_cli.state import get_state

    state = get_state()

    def print_intro() -> None:
        info_text = (
            f"[bold]Welcome to the Zabbix command-line interface (v{__version__})[/]\n"
            f"[bold]Connected to server {state.config.api.url} (v{state.client.version})[/]"
        )
        info_panel = Panel(
            green(info_text),
            expand=False,
            padding=(0, 1),
        )
        help_text = "Type --help to list commands, :h for REPL help, :q to exit."
        intro = Group(info_panel, help_text)
        console.print(intro)

    def pre_run() -> None:
        if not state.repl:
            print_intro()
            state.repl = True
        state.revert_config_overrides()

    prompt_kwargs: dict[str, Any] = {"pre_run": pre_run, "history": state.history}
    start_repl(ctx, app, prompt_kwargs=prompt_kwargs)


def version_callback(value: bool):
    if value:
        print(f"zabbix-cli version {__version__}")
        raise typer.Exit()


@app.callback(invoke_without_command=True)
def main_callback(
    ctx: typer.Context,
    config_file: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        help="Alternate configuration file to use.",
        show_default=False,
    ),
    input_file: Optional[Path] = typer.Option(
        None,
        "--file",
        "--input-file",  # DEPRECATED: V2 name for compatibility
        "-f",
        help="File containing Zabbix-CLI commands to execute in bulk.",
        rich_help_panel="Bulk Mode Options",
        show_default=False,
    ),
    bulk_mode: Optional[BulkRunnerMode] = typer.Option(
        None,
        "--bulk-mode",
        "-b",
        help="Error handling strategy when running in bulk mode.",
        case_sensitive=False,
        rich_help_panel="Bulk Mode Options",
        show_default=False,
    ),
    output_format: Optional[OutputFormat] = typer.Option(
        None,
        "--format",
        "--output-format",  # DEPRECATED: V2 name for compatibility
        "-o",
        help="Define the output format when running in command-line mode.",
        case_sensitive=False,
        show_default=False,
    ),
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        help="Show the version of Zabbix-CLI and exit.",
        is_eager=True,
        callback=version_callback,
    ),
    # Deprecated option, kept for compatibility with V2
    zabbix_command: Optional[str] = typer.Option(
        None,
        "--command",
        "-C",
        help="Zabbix-CLI command to execute when running in command-line mode.",
        hidden=True,
        show_default=False,
    ),
) -> None:
    # Don't run callback if --help is passed in
    # https://github.com/tiangolo/typer/issues/55
    if "--help" in sys.argv:
        return

    state = get_state()
    if not should_skip_configuration(ctx) and not state.is_config_loaded:
        state.config = get_config(config_file, init=True)

    # Config overrides are always applied
    if output_format is not None:
        state.config.app.output.format = output_format
    if bulk_mode is not None:
        state.config.app.bulk_mode = bulk_mode

    if state.repl or state.bulk:
        return  # In REPL or bulk mode already; no need to re-configure.

    logger.debug("Zabbix-CLI started.")

    if should_skip_login(ctx):
        return

    state.login()
    # Configure plugins _after_ login
    # This allows plugins to use the Zabbix API client + more
    # in their __configure__ functions.
    app.configure_plugins(state.config)

    # TODO: look at order of evaluation here. What takes precedence?
    # Should passing both --input-file and --command be an error? probably!
    if zabbix_command:
        from zabbix_cli._v2_compat import run_command_from_option

        run_command_from_option(ctx, zabbix_command)
        return
    elif input_file:
        from zabbix_cli.bulk import run_bulk

        run_bulk(ctx, input_file, state.config.app.bulk_mode)
    elif ctx.invoked_subcommand is not None:
        return  # modern alternative to `-C` option to run a single command
    else:
        # If no command is passed in, we enter the REPL
        run_repl(ctx)


# TODO: Add a decorator for skipping or some sort of parameter to the existing
#       StatefulApp.command method that marks a command as not requiring
#       a configuration file to be loaded.


def should_skip_configuration(ctx: typer.Context) -> bool:
    """Check if the command should skip all configuration of the app."""
    return ctx.invoked_subcommand in [
        "update",
        "open",
        "sample_config",
        "show_dirs",
        "init",
    ]


def should_skip_login(ctx: typer.Context) -> bool:
    """Check if the command should skip logging in to the Zabbix API."""
    if should_skip_configuration(ctx):
        return True
    return ctx.invoked_subcommand in ["migrate_config", "update_config", "show_config"]


def _parse_config_arg() -> Optional[Path]:
    """Get a custom config file path from the command line arguments.

    Modifies sys.argv in place to remove the --config/-c option and its
    argument in order to load the config before instantiating the Typer app.

    This hack enables us to read plugins from the configuration file
    and load them _before_ we call `app()`. This lets commands defined
    in plugins to be used in single-command mode as well as showing up
    when the user types `--help`. Otherwise, the plugin commands would
    not be registered to the active Click command group that drives the CLI.
    """
    opts = ["--config", "-c"]
    for opt in opts:
        if opt in sys.argv:
            index = sys.argv.index(opt)
            break
    else:
        return None

    # If we have the option, we need an argument
    if not len(sys.argv) > index + 1 or not (conf := sys.argv[index + 1].strip()):
        from zabbix_cli.output.console import exit_err

        exit_err("No value provided for --config/-c argument.")
    return Path(conf)


def main() -> int:
    """Main entry point for the CLI."""
    state = get_state()

    try:
        # Load config before launching the app in order to:
        # - configure logging
        # - configure console output
        # - load local plugins (defined in the config)
        configure_logging()
        conf = _parse_config_arg()
        config = get_config(conf, init=False)
        state.config = config
        app.load_plugins(state.config)
        app()
    except Exception as e:
        from zabbix_cli.exceptions import handle_exception

        handle_exception(e)
    finally:
        state.logout_on_exit()
        logger.debug("Zabbix-CLI stopped.")
    return 0


if __name__ == "__main__":
    main()



"""Commands for managing media."""

from __future__ import annotations

import typer

from zabbix_cli.app import app
from zabbix_cli.output.render import render_result

HELP_PANEL = "Media"


@app.command("show_media_types", rich_help_panel=HELP_PANEL)
def show_media_types(ctx: typer.Context) -> None:
    """Show all available media types."""
    from zabbix_cli.models import AggregateResult

    media_types = app.state.client.get_mediatypes()

    render_result(AggregateResult(result=media_types))



from __future__ import annotations


def _pluralize_word(word: str, count: int) -> str:
    if count == 1:
        return word
    if word.endswith("y"):
        return word[:-1] + "ies"
    return word + "s"


def pluralize(word: str, count: int, *, with_count: bool = True) -> str:
    """Pluralize a word based on a count.

    Examples:
    >>> from zabbix_cli.output.formatting.grammar import pluralize as p
    >>> p("apple", 1)
    '1 apple'
    >>> p("apple", 2)
    '2 apples'
    >>> p("category", 1)
    '1 category'
    >>> p("category", 2)
    '2 categories'
    >>> p("category", 0)
    '0 categories'
    >>> p("category", 0, with_count=False) # see pluralize_no_count
    'categories'
    """
    if with_count:
        return f"{count} {_pluralize_word(word, count)}"
    return _pluralize_word(word, count)


def pluralize_no_count(word: str, count: int) -> str:
    """Pluralize a word without a count prepended to the pluralized word.

    Shortcut for `pluralize(word, count, with_count=False)`.

    Examples:
    >>> from zabbix_cli.output.formatting.grammar import pluralize_no_count as pnc
    >>> pnc("apple", 1)
    'apple'
    >>> pnc("apple", 2)
    'apples'
    >>> pnc("category", 1)
    'category'
    >>> pnc("category", 2)
    'categories'
    """
    return _pluralize_word(word, count)



from __future__ import annotations

import sys

if sys.version_info >= (3, 10):
    from types import EllipsisType

    EllipsisType = EllipsisType
else:
    from typing import Any

    EllipsisType = Any



"""Generates a YAML file containing all the global options for the CLI."""

from __future__ import annotations

import logging
import sys
from enum import Enum
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Optional
from typing import Union
from typing import get_args
from typing import get_origin

import tomli_w
import yaml  # type: ignore
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import Json
from pydantic import RootModel
from pydantic import SecretStr
from pydantic import TypeAdapter
from pydantic import ValidationInfo
from pydantic import computed_field
from pydantic import field_validator
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined
from typing_extensions import Self
from zabbix_cli.config.model import Config

sys.path.append(Path(__file__).parent.as_posix())
from common import DATA_DIR  # noqa
from common import add_path_placeholders  # noqa


JSONAdapter = TypeAdapter(Json)

TYPE_MAP = {
    SecretStr: "str",
    Path: "str",
    Literal: "str",
}
"""Special types that are represented differently in config file and in the code."""


TYPE_CAN_STR = {str, int, float, bool, list, dict, set, tuple, type(None)}
"""Types that can be represented by calling str() on them"""
# NOTE: Does this apply to all built-ins? Can we just check for builtins?


# HACK: dict retrieval with type hinting
def get_field_info(info: ValidationInfo) -> FieldInfo:
    return info.data["field"]


class ConfigBase(BaseModel):
    """Common fields shared by config tables and options."""

    field: Optional[FieldInfo] = Field(default=None, exclude=True)

    name: str
    description: str = ""
    parents: list[str] = []

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @computed_field()
    def is_model(self) -> bool:
        return hasattr(self, "fields")

    @field_validator("description", mode="before")
    @classmethod
    def validate_description(cls, value: Any) -> str:
        return value or ""

    @field_validator("description", mode="after")
    @classmethod
    def dedent_description(cls, value: str) -> str:
        return "\n".join(line.strip() for line in value.splitlines())


class ConfigOption(ConfigBase):
    type: str
    default: Any = None
    required: bool = False
    examples: Optional[list[Any]] = None

    @computed_field()
    def is_model(self) -> bool:
        return False

    @computed_field()
    @property
    def choices(self) -> Optional[list[Any]]:
        # Handle common choice types
        if not self.field or self.field.annotation is None:
            return None
        origin = get_origin(self.field.annotation)
        if origin is Literal:
            return list(get_args(self.field.annotation))
        elif lenient_issubclass(self.field.annotation, Enum):
            return list(self.field.annotation)
        return None

    @computed_field()
    @property
    def choices_str(self) -> Optional[str]:
        if not self.choices:
            return None
        return ", ".join(str(choice) for choice in self.choices)

    @computed_field()
    @property
    def parents_str(self) -> str:
        return ".".join(self.parents)

    def example_toml_dict(self) -> dict[str, Any]:
        if not self.examples:
            # We have no examples to provide, this is a problem
            raise ValueError(
                f"Cannot render field {self.name!r}. "
                "It has no defaults and no examples. "
                "Provide an example in the field definition under `examples`."
            )
        example = self.examples[0]

        ex: dict[str, Any] = {}
        current = ex
        if self.parents:
            for parent in self.parents:
                current[parent] = {}
                current = current[parent]
        current[self.name] = example
        return ex

    @computed_field()
    @property
    def example(self) -> str:
        """TOML representation of the first example."""
        ex = self.example_toml_dict()
        ex_jsonable = JSONAdapter.dump_python(ex, exclude_none=True, mode="json")
        return tomli_w.dumps(ex_jsonable)

    @classmethod
    def from_field_info(
        cls, name: str, field: FieldInfo, parents: list[str]
    ) -> ConfigOption:
        return cls(
            # WARNING: DO NOT CHANGE THE ORDER OF THE `field` PARAMETER
            # `field` must be validated first in order to have access to
            # the field data in the validation methods
            field=field,
            # Custom param to tell where we are in the model hierarchy
            parents=parents,
            # Rest of the parameters
            name=name,
            type=field.annotation,  # type: ignore # field validator
            description=field.description,  # type: ignore # field validator
            default=field.default,
            required=field.default is PydanticUndefined and not field.default_factory,
            examples=field.examples,
        )

    @field_validator("default", mode="before")
    @classmethod
    def validate_default(cls, value: Any) -> Optional[Any]:
        if value is PydanticUndefined:
            return None
        if isinstance(value, SecretStr):
            return value.get_secret_value()
        if isinstance(value, bool):
            return str(value).lower()
        return value

    @field_validator("type", mode="before")
    @classmethod
    def validate_type(cls, value: Any) -> str:
        if value is None:
            return "Any"

        origin = get_origin(value)
        args = get_args(value)

        def type_to_str(t: type[Any]) -> str:
            if lenient_issubclass(value, str):
                return "str"

            if lenient_issubclass(value, Enum):
                # Get the name of the first enum member type
                # NOTE: Will fail if enum has no members
                return str(list(value)[0])  # pyright: ignore[reportUnknownArgumentType]

            # Types that are represented as strings in config (paths, secrets, etc.)
            if typ := TYPE_MAP.get(t):
                return typ

            # Primitives and built-in generics (str, int, list[str], dict[str, int], etc.)
            if origin in TYPE_CAN_STR:
                return str(value)

            # Fall back on the string representation of the type
            return getattr(value, "__name__", str(value))

        # Handle generics, literals, etc.
        if origin and args:
            # Get the name of the first type in the Literal type
            # NOTE: we expect that Literal is only used with a single type
            if origin is Literal:
                return args[0].__class__.__name__
            # Get first non-None type in Union
            # NOTE: we expect that the config does not have unions of more than 2 types
            elif origin is Union and args:
                # Strip None from the Union
                ar = (type_to_str(a) for a in args if a is not type(None))
                return " | ".join(ar)

        return type_to_str(value)

    @field_validator("examples", mode="before")
    @classmethod
    def validate_examples(
        cls, value: Optional[list[Any]], info: ValidationInfo
    ) -> list[Any]:
        if value:
            return value
        field = get_field_info(info)
        if field.default is not PydanticUndefined:
            return [field.default]
        elif field.default_factory:
            return [field.default_factory()]
        return []


class ConfigTable(ConfigBase):
    # NOTE: can we generalize this to always be a list of ConfigOption?
    # Can we get rid of ConfigTable altogether and just compose everything of
    # ConfigOption by adding `fields` to the ConfigOption model?
    # That way we could have a consistent interface regardless of
    # whether we're dealing with a submodel or a field.
    fields: list[Union[ConfigTable, ConfigOption]]

    @classmethod
    def from_field_info(
        cls, name: str, field: FieldInfo, parents: list[str], field_parents: list[str]
    ) -> Self:
        assert field.annotation
        return cls(
            field=field,
            parents=parents,
            name=name,
            description=field.annotation.__doc__,  # type: ignore # validator
            fields=get_config_options(field.annotation, name, field_parents),
        )

    def example_toml_dict(self) -> dict[str, Any]:
        return {}  # HACK: avoid isinstance checking

    @computed_field()
    @property
    def example(self) -> str:
        ex: dict[str, Any] = {}
        # Pretty stupid way to render an example, but it works
        for field in self.fields:
            if field.is_model:
                continue
            example = field.example_toml_dict()
            if not ex:
                ex.update(example)
            else:
                if self.name in ex:
                    ex[self.name].update(example)
                elif self.parents:
                    e = ex
                    for parent in self.parents:
                        if not e.get(parent):
                            e[parent] = {}
                        e = e[parent]
                    e[self.name.rpartition(".")[-1]].update(example)
                    ex.update(e)
                else:
                    ex.update(example)

        ex_jsonable = JSONAdapter.dump_python(ex, exclude_none=True, mode="json")
        return tomli_w.dumps(ex_jsonable)


def lenient_issubclass(cls: type, class_or_tuple: Union[type, tuple[type]]) -> bool:
    try:
        return issubclass(cls, class_or_tuple)
    except TypeError:
        return False


def get_config_options(
    type_: type[BaseModel], current_name: str = "", parents: Optional[list[str]] = None
) -> list[Union[ConfigTable, ConfigOption]]:
    """Recursively extract the configuration options from a Pydantic model."""
    if parents is None:
        parents = []
    options: list[Union[ConfigTable, ConfigOption]] = []
    for field_name, field in type_.model_fields.items():
        if field.exclude:
            continue
        if not field.annotation:
            continue
        if field.default is None:
            continue
        if field.deprecated:
            continue

        if lenient_issubclass(field.annotation, RootModel):
            logging.debug("Skipping %s. It is a root model.", field_name)
            continue

        if current_name:
            name = f"{current_name}.{field_name}"
        else:
            name = field_name

        if lenient_issubclass(field.annotation, BaseModel):
            # We have a nested model
            field_parents = parents.copy()
            field_parents.append(field_name)
            options.append(
                ConfigTable.from_field_info(name, field, parents, field_parents)
            )
        else:
            # We have a field
            options.append(
                ConfigOption.from_field_info(field_name, field, parents=parents)
            )
    return options


def generate_config_info() -> ConfigTable:
    """Generate the configuration options for the CLI."""
    conf = ConfigTable(name="", fields=[])
    conf.fields = get_config_options(Config, conf.name)
    return conf


def main() -> None:
    conf = generate_config_info()
    out = yaml.dump(conf.model_dump(mode="json", exclude_none=True), sort_keys=False)
    out = add_path_placeholders(out)  # type: ignore
    # Replace paths with placeholders
    with open(DATA_DIR / "config_options.yaml", "w") as f:
        f.write(out)


if __name__ == "__main__":
    main()



"""Commands for managing users."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Optional
from typing import TypeVar

import typer
from strenum import StrEnum

from zabbix_cli._v2_compat import ARGS_POSITIONAL
from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.exceptions import ZabbixNotFoundError
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import info
from zabbix_cli.output.console import success
from zabbix_cli.output.console import warning
from zabbix_cli.output.prompts import str_prompt
from zabbix_cli.output.render import render_result
from zabbix_cli.pyzabbix.enums import UserRole
from zabbix_cli.utils.args import check_at_least_one_option_set
from zabbix_cli.utils.args import parse_bool_arg
from zabbix_cli.utils.args import parse_list_arg

if TYPE_CHECKING:
    from typing import Protocol

    class UsergroupLike(Protocol):
        name: str
        usrgrpid: str

    UsergroupLikeT = TypeVar("UsergroupLikeT", bound=UsergroupLike)


HELP_PANEL = "User"


def get_random_password() -> str:
    import hashlib
    import random

    x = hashlib.md5()
    x.update(str(random.randint(1, 1000000)).encode("ascii"))
    return x.hexdigest()


@app.command(
    "create_user",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Create a user with a random password",
            "create_user jdoe --firstname John --lastname Doe",
        ),
        Example(
            "Create a user with a specific password",
            "create_user jdoe --firstname John --lastname Doe --passwd 'MyPassword'",
        ),
        Example(
            "Create a user with a password from stdin",
            "create_user jdoe --firstname John --lastname Doe --passwd -",
        ),
        Example(
            "Create a user with a specific role and user groups",
            "create_user jdoe --firstname John --lastname Doe --role admin --groups 'All-admins,Linux-admins'",
        ),
    ],
)
def create_user(
    ctx: typer.Context,
    username: str = typer.Argument(
        help="Username of the user to create.",
        show_default=False,
    ),
    first_name: Optional[str] = typer.Option(
        None, "--firstname", help="First name of the user to create."
    ),
    last_name: Optional[str] = typer.Option(
        None, "--lastname", "--surname", help="Last name of the user to create."
    ),
    password: Optional[str] = typer.Option(
        None,
        "--passwd",
        "--password",
        help="Password of the user to create. Set to '-' to prompt for password. Generates random password if omitted.",
    ),
    role: UserRole = typer.Option(
        UserRole.USER,
        "--role",
        help="Role of the user.",
        case_sensitive=False,
    ),
    autologin: bool = typer.Option(False, help="Enable auto-login for the user."),
    autologout: str = typer.Option(
        "86400",
        help="User session lifetime in seconds. Set to 0 to never expire. Can be a time unit with suffix (0s, 15m, 1h, 1d, etc.)",
    ),
    groups: Optional[str] = typer.Option(
        None,
        "--groups",
        "--usergroups",
        help="Comma-separated list of user group names or IDs to add the user to.",
        show_default=False,
    ),
    use_default_usergroups: bool = typer.Option(
        True,
        "--default-usergroups/--no-default-usergroups",
        help=(
            "Add the user to the default user groups defined in the config file in addition to "
            "any groups specified with [option]--groups[/]. This is the default behavior."
        ),
        show_default=True,
    ),
    # Legacy V2 positional args
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Create a user."""
    from zabbix_cli.models import Result
    from zabbix_cli.pyzabbix.types import User

    try:
        app.state.client.get_user(username)
        exit_err(f"User {username!r} already exists.")
    except ZabbixNotFoundError:
        pass

    if args:
        # Old args format: <username>  <first_name> <last_name> <password> <type> <autologin> <autologout> <usergroups>
        # We already have username, so we are left with 7 args.
        # In V2, we either expected NO positional args or ALL of them.
        # So we just match that behavior here.
        if len(args) != 7:
            exit_err(
                "Invalid number of positional arguments. Please use options instead."
            )
        first_name = args[0]
        last_name = args[1]
        password = args[2]
        role = UserRole(args[3])
        autologin = parse_bool_arg(args[4])
        autologout = args[5]
        groups = args[6]

    if password == "-":
        password = str_prompt("Password", password=True)
    elif not password:
        # Generate random password
        password = get_random_password()

    grouplist = parse_list_arg(groups)
    if use_default_usergroups:
        grouplist.extend(app.state.config.app.commands.create_user.usergroups)
    ugroups = [app.state.client.get_usergroup(ug) for ug in set(grouplist)]

    userid = app.state.client.create_user(
        username,
        password,
        first_name=first_name,
        last_name=last_name,
        role=role,
        autologin=autologin,
        autologout=autologout,
        usergroups=ugroups,
    )
    render_result(
        Result(
            message=f"Created user {username!r} ({userid}).",
            result=User(userid=str(userid), username=username),
        ),
    )


@app.command(
    "create_notification_user",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Create a notification user for email reporting",
            "create_notification_user user@example.com Email",
        ),
    ],
)
def create_notification_user(
    ctx: typer.Context,
    sendto: str = typer.Argument(
        help="Email address, SMS number, jabber address, etc.",
        show_default=False,
    ),
    mediatype: str = typer.Argument(
        help="A media type name or ID defined in Zabbix. Case-sensitive.",
        show_default=False,
    ),
    remarks: Optional[str] = typer.Option(
        None,
        "--remarks",
        help="Remarks about the notification user to include in username (max 20 chars).",
    ),
    username: Optional[str] = typer.Option(
        None,
        "--username",
        help="Override generated username. Ignores --remarks.",
    ),
    usergroups: Optional[str] = typer.Option(
        None,
        "--usergroups",
        help="Comma-separated list of usergroups to add the user to. Overrides user groups in config file.",
    ),
    use_default_usergroups: bool = typer.Option(
        True,
        "--default-usergroups/--no-default-usergroups",
        help=(
            "Add the user to the default user groups defined in the config file in addition to "
            "any groups specified with [option]--usergroups[/]. This is the default behavior."
        ),
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Do not create the user, just show what would be done.",
    ),
    # Legacy V2 args
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    # TODO: Improve phrasing of this help text. "Defining media for usergroup"???
    """Create a notification user.

    Notification users can be used to send notifications when a Zabbix
    event happens.

    Sometimes we need to send a notification to a place not owned by any
    user in particular, e.g. an email list or jabber channel but Zabbix does
    not provide a way to define a media for a user group.

    This is the reason we use [i]notification users[/]. They are users nobody
    owns, but that can be used by other users to send notifications to the
    media defined in the notification user profile.

    Run [command]show_media_types[/command] to get a list of available media types.

    Adds the user to the default user group and default notification user group
    unless [option]--usergroups[/option] is specified.
    """
    from zabbix_cli.commands.results.user import CreateNotificationUserResult
    from zabbix_cli.pyzabbix.types import UserMedia

    if args:
        # Old args format: <sendto> <mediatype> <remarks>
        # We already have sendto and mediatype, so we are left with 1 arg.
        if len(args) != 1:
            exit_err(
                "Invalid number of positional arguments. Please use options instead."
            )
        remarks = args[0]
    remarks = remarks or ""

    # Generate username
    if username and remarks:
        warning("Both --username and --remarks specified. Ignoring --remarks.")

    username = get_notification_user_username(username, sendto, remarks)

    # Check if user exists (it should not)
    try:
        app.state.client.get_user(username)
        exit_err(f"User {username!r} already exists.")
    except ZabbixNotFoundError:
        pass

    # Check if media type exists (it should)
    try:
        mt = app.state.client.get_mediatype(mediatype)
    except ZabbixNotFoundError:
        exit_err(
            f"Media type {mediatype!r} does not exist. Run [command]show_media_types[/command] command to get a list of available media types."
        )

    # Parse user groups args or use defaults
    ug_list = parse_list_arg(usergroups)
    if use_default_usergroups:
        ug_list.extend(
            app.state.config.app.commands.create_notification_user.usergroups
        )

    with app.status("Fetching user group(s)..."):
        ugroups = [app.state.client.get_usergroup(ug) for ug in set(ug_list)]

    user_media = [
        UserMedia(
            mediatypeid=mt.mediatypeid,
            sendto=sendto,
            active=0,  # enabled
            severity=63,  # all
            period="1-7,00:00-24:00",  # 24/7
        )
    ]

    result = CreateNotificationUserResult(
        username=username, media=user_media, userid="", usergroups=ugroups
    )
    if dryrun:
        info("Would create:")
        render_result(result)
        return

    with app.status("Creating user..."):
        userid = app.state.client.create_user(
            username=username,
            password=get_random_password(),
            role=UserRole.USER,
            autologin=False,
            autologout="3600",
            usergroups=ugroups,
            media=user_media,
        )

    result.userid = userid

    success(f"Created notification user {username!r} ({userid}).")
    render_result(result)


@app.command("remove_user", rich_help_panel=HELP_PANEL)
def remove_user(
    ctx: typer.Context,
    username: str = typer.Argument(
        help="Username to remove.",
        show_default=False,
    ),
) -> None:
    """Remove a user."""
    from zabbix_cli.models import Result

    user = app.state.client.get_user(username)
    app.state.client.delete_user(user)
    render_result(
        Result(
            message=f"Deleted user {user.name!r} ({user.userid}).",
            result=user,
        ),
    )


@app.command("show_user", rich_help_panel=HELP_PANEL)
def show_user(
    ctx: typer.Context,
    username: str = typer.Argument(
        help="Username of user",
        show_default=False,
    ),
) -> None:
    """Show a user."""
    user = app.state.client.get_user(username)
    render_result(user)


class UserSorting(StrEnum):
    NAME = "name"
    ID = "id"
    ROLE = "role"


@app.command("show_users", rich_help_panel=HELP_PANEL)
def show_users(
    ctx: typer.Context,
    username_or_id: Optional[str] = typer.Argument(
        None,
        help="Filter by username or ID. Supports wildcards.",
        show_default=False,
    ),
    role: Optional[UserRole] = typer.Option(
        None,
        "--role",
        help="Filter by role.",
        case_sensitive=False,
    ),
    limit: Optional[int] = typer.Option(
        None, "--limit", help="Limit the number of users shown."
    ),
    sort: Optional[UserSorting] = typer.Option(
        UserSorting.NAME,
        "--sort",
        help="Sort by field.",
    ),
) -> None:
    """Show all users.

    Users can be filtered by name, ID, or role.
    """
    from zabbix_cli.models import AggregateResult
    from zabbix_cli.pyzabbix.compat import user_name

    us = parse_list_arg(username_or_id)

    # TODO: move this to the client somehow
    #       This is also clumsy, because we want users to pass in
    #       "name", "id", or "role" as arguments to the command,
    #       but the API expects "userid", "name", or "roleid"
    sorting = None
    if sort:
        if sort == UserSorting.ROLE:
            sorting = "roleid"
        elif sort == UserSorting.ID:
            sorting = "userid"
        else:
            sorting = user_name(app.state.client.version)

    with app.status("Fetching users..."):
        users = app.state.client.get_users(
            *us, role=role, limit=limit, sort_field=sorting, sort_order="ASC"
        )
    render_result(AggregateResult(result=users))


def get_notification_user_username(
    username: Optional[str], sendto: str, remarks: str
) -> str:
    """Generate a username for a notification user."""
    username = username.strip().replace(" ", "_") if username else ""
    remarks = remarks.strip()[:20].replace(" ", "_")
    sendto = sendto.strip().replace(".", "-")
    if username:
        return username
    username = "notification-user"
    if remarks:
        username += f"-{remarks}"
    return f"{username}-{sendto}"


@app.command(
    "update_user",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Assign new first and last name",
            "update_user jdoe --firstname John --lastname Doe",
        ),
        Example(
            "Promote user to admin",
            "update_user jdoe --role Admin",
        ),
        Example(
            "Update user's password, prompt for passwords",
            "update_user jdoe --passwd - --old-passwd -",
        ),
        Example(
            "Update user's password, generate random new password",
            "update_user jdoe --passwd ? --old-passwd -",
        ),
        Example(
            "Disable autologin for user",
            "update_user jdoe --no-autologin",
        ),
    ],
)
def update_user(
    ctx: typer.Context,
    username: str = typer.Argument(
        help="Username of the user to update",
        show_default=False,
    ),
    first_name: Optional[str] = typer.Option(
        None, "--firstname", help="New first name."
    ),
    last_name: Optional[str] = typer.Option(None, "--lastname", help="New last name."),
    new_password: Optional[str] = typer.Option(
        None,
        "--passwd",
        "--new-passwd",
        help="New password for user. Set to '-' to prompt for password, '?' to generate a random password.",
    ),
    old_password: Optional[str] = typer.Option(
        None,
        "--old-passwd",
        help="Existing password, required if --passwd is used. Set to '-' to prompt for password.",
    ),
    role: Optional[UserRole] = typer.Option(
        None,
        "--role",
        help="User role.",
        case_sensitive=False,
    ),
    autologin: Optional[bool] = typer.Option(
        None,
        "--autologin/--no-autologin",
        help="Enable/disable auto-login",
        show_default=False,
    ),
    autologout: Optional[str] = typer.Option(
        None,
        help="User session lifetime in seconds. Set to 0 to never expire. Can be a time unit with suffix (0s, 15m, 1h, 1d, etc.)",
    ),
    # Legacy V2 positional args
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Update a user.

    Use [command]add_user_to_usergroup[/command] and [command]remove_user_from_usergroup[/command] to manage user groups.
    """
    from zabbix_cli.models import Result

    check_at_least_one_option_set(ctx)

    user = app.state.client.get_user(username)

    if new_password == "-":
        new_password = str_prompt("New password", password=True)
    elif new_password == "?":
        new_password = get_random_password()

    if new_password:
        if not old_password:
            exit_err("Old password is required when changing password.")
        if old_password == "-":
            old_password = str_prompt("Old password", password=True)

    app.state.client.update_user(
        user,
        current_password=old_password,
        new_password=new_password,
        first_name=first_name,
        last_name=last_name,
        role=role,
        autologin=autologin,
        autologout=autologout,
    )
    render_result(
        Result(
            message=f"Updated user {user.name!r} ({user.userid}).",
            result=user,
        ),
    )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Optional

from pydantic import Field
from pydantic import model_serializer
from typing_extensions import Self

from zabbix_cli.models import TableRenderable
from zabbix_cli.pyzabbix.enums import MacroAutomatic
from zabbix_cli.pyzabbix.types import Macro

if TYPE_CHECKING:
    from zabbix_cli.models import ColsRowsType
    from zabbix_cli.models import RowsType


class ShowHostUserMacrosResult(TableRenderable):
    hostmacroid: str
    macro: str
    value: Optional[str] = None
    type: str
    description: Optional[str] = None
    hostid: str
    automatic: Optional[int] = None

    @classmethod
    def from_result(cls, macro: Macro) -> Self:
        return cls(
            hostmacroid=macro.hostmacroid,
            macro=macro.macro,
            value=macro.value,
            type=macro.type_str,
            description=macro.description,
            hostid=macro.hostid,
            automatic=macro.automatic,
        )

    def __cols_rows__(self) -> ColsRowsType:
        cols = [
            "Macro ID",
            "Macro",
            "Value",
            "Type",
            "Description",
            "Host ID",
            "Automatic",
        ]
        rows: RowsType = [
            [
                self.hostmacroid,
                self.macro,
                str(self.value),
                self.type,
                self.description or "",
                self.hostid,
                MacroAutomatic.string_from_value(self.automatic),
            ]
        ]
        return cols, rows


class MacroHostListV2(TableRenderable):
    macro: Macro

    def __cols_rows__(self) -> ColsRowsType:
        rows: RowsType = [
            [self.macro.macro, str(self.macro.value), host.hostid, host.host]
            for host in self.macro.hosts
        ]
        return ["Macro", "Value", "HostID", "Host"], rows

    @model_serializer()
    def model_ser(self) -> dict[str, Any]:
        if not self.macro.hosts:
            return {}  # match V2 output
        return {
            "macro": self.macro.macro,
            "value": self.macro.value,
            "hostid": self.macro.hosts[0].hostid,
            "host": self.macro.hosts[0].host,
        }


class MacroHostListV3(TableRenderable):
    macro: Macro

    def __cols_rows__(self) -> ColsRowsType:
        rows: RowsType = [
            [host.hostid, host.host, self.macro.macro, str(self.macro.value)]
            for host in self.macro.hosts
        ]
        return ["Host ID", "Host", "Macro", "Value"], rows


class GlobalMacroResult(TableRenderable):
    """Result of `define_global_macro` command."""

    globalmacroid: str = Field(json_schema_extra={"header": "Global Macro ID"})
    macro: str
    value: Optional[str] = None  # for usermacro.get calls


class ShowUsermacroTemplateListResult(TableRenderable):
    macro: str
    value: Optional[str] = None
    templateid: str
    template: str

    def __cols__(self) -> list[str]:
        return ["Macro", "Value", "Template ID", "Template"]



"""Simple in-memory caching of frequently used Zabbix objects."""

# TODO: add on/off toggle for caching
from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from typing import Optional

from zabbix_cli.exceptions import ZabbixCLIError

if TYPE_CHECKING:
    from zabbix_cli.pyzabbix.client import ZabbixAPI


class ZabbixCache:
    """In-memory cache of frequently used Zabbix objects."""

    def __init__(self, client: ZabbixAPI) -> None:
        self.client = client
        self._hostgroup_name_cache: dict[str, str] = {}
        """Mapping of hostgroup names to hostgroup IDs"""

        self._hostgroup_id_cache: dict[str, str] = {}
        """Mapping of hostgroup IDs to hostgroup names"""

        self._templategroup_name_cache: dict[str, str] = {}
        """Mapping of templategroup names to templategroup IDs"""

        self._templategroup_id_cache: dict[str, str] = {}  # NOTE: unused
        """Mapping of templategroup IDs to templategroup names"""

    def populate(self) -> None:
        try:
            self._populate_hostgroup_cache()
            self._populate_templategroup_cache()
        except Exception as e:
            raise ZabbixCLIError(f"Failed to populate Zabbix cache: {e}") from e

    def _populate_hostgroup_cache(self) -> None:
        """Populates the hostgroup caches with data from the Zabbix API."""
        hostgroups = self.client.hostgroup.get(output=["name", "groupid"])
        self._hostgroup_name_cache = {
            hostgroup["name"]: hostgroup["groupid"] for hostgroup in hostgroups
        }
        self._hostgroup_id_cache = {
            hostgroup["groupid"]: hostgroup["name"] for hostgroup in hostgroups
        }

    def _populate_templategroup_cache(self) -> None:
        """Populates the templategroup caches with data from the Zabbix API
        on Zabbix >= 6.2.0.
        """
        if self.client.version.release < (6, 2, 0):
            logging.debug(
                "Skipping template group caching. API version is %s",
                self.client.version,
            )
            return

        templategroups = self.client.templategroup.get(output=["name", "groupid"])
        self._templategroup_name_cache = {
            templategroup["name"]: templategroup["groupid"]
            for templategroup in templategroups
        }
        self._templategroup_id_cache = {
            templategroup["groupid"]: templategroup["name"]
            for templategroup in templategroups
        }

    def get_hostgroup_name(self, hostgroup_id: str) -> Optional[str]:
        """Returns the name of a host group given its ID."""
        return self._hostgroup_id_cache.get(hostgroup_id)

    def get_hostgroup_id(self, hostgroup_name: str) -> Optional[str]:
        """Returns the ID of a host group given its name."""
        return self._hostgroup_name_cache.get(hostgroup_name)

    def get_templategroup_name(self, templategroup_id: str) -> Optional[str]:
        """Returns the name of a template group given its ID."""
        return self._templategroup_id_cache.get(templategroup_id)

    def get_templategroup_id(self, templategroup_name: str) -> Optional[str]:
        """Returns the ID of a template group given its name."""
        return self._templategroup_name_cache.get(templategroup_name)



from __future__ import annotations

from collections.abc import Generator
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Optional
from unittest.mock import patch

import pytest
from inline_snapshot import snapshot
from packaging.version import Version
from pydantic import SecretStr
from pytest_httpserver import HTTPServer
from zabbix_cli import auth
from zabbix_cli._v2_compat import AUTH_FILE as AUTH_FILE_LEGACY
from zabbix_cli._v2_compat import AUTH_TOKEN_FILE as AUTH_TOKEN_FILE_LEGACY
from zabbix_cli.auth import Authenticator
from zabbix_cli.auth import Credentials
from zabbix_cli.auth import CredentialsSource
from zabbix_cli.auth import CredentialsType
from zabbix_cli.auth import SessionFile
from zabbix_cli.auth import SessionInfo
from zabbix_cli.auth import SessionList
from zabbix_cli.auth import get_auth_file_paths
from zabbix_cli.auth import get_auth_token_file_paths
from zabbix_cli.config.constants import AUTH_FILE
from zabbix_cli.config.constants import AUTH_TOKEN_FILE
from zabbix_cli.config.model import Config
from zabbix_cli.exceptions import SessionFileError
from zabbix_cli.exceptions import SessionFileNotFoundError
from zabbix_cli.exceptions import SessionFilePermissionsError
from zabbix_cli.pyzabbix.client import ZabbixAPI

from tests.utils import add_zabbix_endpoint
from tests.utils import add_zabbix_version_endpoint

if TYPE_CHECKING:
    from zabbix_cli.models import TableRenderable


def test_get_auth_file_paths_defafult(config: Config) -> None:
    """Test the default auth file paths."""
    # Path from config is not present (same as default AUTH_FILE)
    assert get_auth_file_paths(config) == [
        AUTH_FILE,
        AUTH_FILE_LEGACY,
    ]


def test_get_auth_file_paths_override(tmp_path: Path, config: Config) -> None:
    """Test overriding the default auth file path in the config."""
    auth_file = tmp_path / "auth"
    config.app.auth_file = auth_file
    # Path from config is first (highest priority)
    assert get_auth_file_paths(config) == [
        auth_file,
        AUTH_FILE,
        AUTH_FILE_LEGACY,
    ]


def test_get_auth_token_file_paths_defafult(config: Config) -> None:
    """Test the default auth token file paths."""
    # Path from config is not present (same as default AUTH_TOKEN_FILE)
    assert get_auth_token_file_paths(config) == [
        AUTH_TOKEN_FILE,
        AUTH_TOKEN_FILE_LEGACY,
    ]


def test_get_auth_token_file_paths_override(tmp_path: Path, config: Config) -> None:
    """Override the default auth token file path in the config."""
    auth_file = tmp_path / "auth_token"
    config.app.auth_token_file = auth_file
    # Path from config is first (highest priority)
    assert get_auth_token_file_paths(config) == [
        auth_file,
        AUTH_TOKEN_FILE,
        AUTH_TOKEN_FILE_LEGACY,
    ]


@pytest.fixture
def table_renderable_mock(monkeypatch) -> type[TableRenderable]:
    """Replace TableRenderable class in zabbix_cli.models with mock class
    so that tests can mutate it without affecting other tests.
    """
    from zabbix_cli.models import TableRenderable

    class MockTableRenderable(TableRenderable):
        pass

    # Set a default version that is different from the real class
    # so that we can tell if the mock class was used
    MockTableRenderable.zabbix_version = Version("0.0.1")

    # Use monkeypatch to temporarily replace the real class with the mock class
    monkeypatch.setattr("zabbix_cli.models.TableRenderable", MockTableRenderable)

    return MockTableRenderable


@pytest.fixture(name="session_file")
def _session_file(tmp_path: Path) -> Generator[Path, None, None]:
    yield tmp_path / ".zabbix-cli_session.json"


@pytest.fixture(name="auth_token_file")
def _auth_token_file(tmp_path: Path) -> Generator[Path, None, None]:
    yield tmp_path / ".zabbix-cli_auth_token"


@pytest.fixture(name="auth_file")
def _auth_file(tmp_path: Path) -> Generator[Path, None, None]:
    yield tmp_path / ".zabbix-cli_auth"


@pytest.mark.parametrize(
    "sources, expect_type, expect_source",
    [
        pytest.param(
            [
                (CredentialsType.AUTH_TOKEN, CredentialsSource.ENV),
                (CredentialsType.AUTH_TOKEN, CredentialsSource.CONFIG),
                (CredentialsType.SESSION, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.CONFIG),
                (CredentialsType.PASSWORD, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.ENV),
                (CredentialsType.AUTH_TOKEN, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.PROMPT),
            ],
            CredentialsType.AUTH_TOKEN,
            CredentialsSource.ENV,
            id="expect_auth_token_env",
        ),
        pytest.param(
            [
                (CredentialsType.AUTH_TOKEN, CredentialsSource.CONFIG),
                (CredentialsType.SESSION, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.CONFIG),
                (CredentialsType.PASSWORD, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.ENV),
                (CredentialsType.AUTH_TOKEN, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.PROMPT),
            ],
            CredentialsType.AUTH_TOKEN,
            CredentialsSource.CONFIG,
            id="expect_auth_token_config",
        ),
        pytest.param(
            [
                (CredentialsType.SESSION, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.CONFIG),
                (CredentialsType.PASSWORD, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.ENV),
                (CredentialsType.AUTH_TOKEN, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.PROMPT),
            ],
            CredentialsType.SESSION,
            CredentialsSource.FILE,
            id="expect_session_file",
        ),
        pytest.param(
            [
                (CredentialsType.PASSWORD, CredentialsSource.ENV),
                (CredentialsType.PASSWORD, CredentialsSource.CONFIG),
                (CredentialsType.PASSWORD, CredentialsSource.FILE),
                (CredentialsType.AUTH_TOKEN, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.PROMPT),
            ],
            CredentialsType.PASSWORD,
            CredentialsSource.ENV,
            id="expect_password_env",
        ),
        pytest.param(
            [
                (CredentialsType.PASSWORD, CredentialsSource.CONFIG),
                (CredentialsType.PASSWORD, CredentialsSource.FILE),
                (CredentialsType.AUTH_TOKEN, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.PROMPT),
            ],
            CredentialsType.PASSWORD,
            CredentialsSource.CONFIG,
            id="expect_password_config",
        ),
        pytest.param(
            [
                (CredentialsType.PASSWORD, CredentialsSource.FILE),
                (CredentialsType.AUTH_TOKEN, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.PROMPT),
            ],
            CredentialsType.PASSWORD,
            CredentialsSource.FILE,
            id="expect_password_file",
        ),
        pytest.param(
            [
                (CredentialsType.AUTH_TOKEN, CredentialsSource.FILE),
                (CredentialsType.PASSWORD, CredentialsSource.PROMPT),
            ],
            CredentialsType.SESSION,
            CredentialsSource.FILE,
            id="expect_legacy_auth_token_file",
        ),
        pytest.param(
            [
                (CredentialsType.PASSWORD, CredentialsSource.PROMPT),
            ],
            CredentialsType.PASSWORD,
            CredentialsSource.PROMPT,
            id="expect_password_prompt",
        ),
    ],
)
def test_authenticator_login_with_any(
    monkeypatch: pytest.MonkeyPatch,
    httpserver: HTTPServer,
    table_renderable_mock: type[TableRenderable],
    auth_token_file: Path,
    auth_file: Path,
    session_file: Path,
    config: Config,
    sources: list[tuple[CredentialsType, CredentialsSource]],
    expect_source: CredentialsSource,
    expect_type: CredentialsType,
) -> None:
    """Test that automatic selection of authentication source works given
    multiple valid authentication sources.


    Mocks ZabbixAPI methods to avoid reliace on a Zabbix server.
    """
    # TODO: test with a mix of valid and invalid sources, as well as
    #  with no valid sources.
    MOCK_USER = "Admin"
    MOCK_PASSWORD = "zabbix"
    MOCK_TOKEN = "abc1234567890"
    # Mock token used in legacy auth token file (to differentiate from new location)
    MOCK_URL = httpserver.url_for("/")

    config.api.url = MOCK_URL
    config.app.auth_token_file = auth_token_file
    config.app.session_file = session_file

    authenticator = Authenticator(config)

    # Endpoint for ZabbixAPI.version
    # Add a version different from the default so we can tell if the mock was used
    # But also is recent enough that we use auth headers (>= 6.4.0)
    add_zabbix_version_endpoint(httpserver, "6.4.0", id=0)

    # Endpoint for ZabbixAPI.login()
    add_zabbix_endpoint(
        httpserver,
        method="user.login",
        params={},  # matching NYI
        response=MOCK_TOKEN,
        id=1,
    )

    # Replace certain difficult to test methods with mocks
    # States reasons for mocking each method

    # REASON: Implementation detail, not relevant to this test.
    # This method is called by the login method as of 3.4.2, but
    # it is an implementation detail and not relevant to this test
    # (it calls host.get with a limit of 1)
    def mock_ensure_authenticated(self: ZabbixAPI) -> None:
        return

    monkeypatch.setattr(
        auth.ZabbixAPI, "ensure_authenticated", mock_ensure_authenticated
    )

    # REASON: Prompts for input (could also be tested with a fake input stream)
    def mock_get_username_password_prompt() -> Credentials:
        return Credentials(
            username=MOCK_USER,
            password=MOCK_PASSWORD,
            source=CredentialsSource.PROMPT,
        )

    monkeypatch.setattr(
        authenticator,
        "_get_username_password_prompt",
        mock_get_username_password_prompt,
    )

    # REASON: Falls back on default auth token file path (which might exist on test user's system)
    # We want to ensure that the function only finds the test file we created
    def mock_get_auth_token_file_paths(config: Optional[Config] = None) -> list[Path]:
        return [auth_token_file]

    monkeypatch.setattr(
        auth, "get_auth_token_file_paths", mock_get_auth_token_file_paths
    )

    # REASON: Same as above
    def mock_get_auth_file_paths(config: Optional[Config] = None) -> list[Path]:
        return [auth_file]

    monkeypatch.setattr(auth, "get_auth_file_paths", mock_get_auth_file_paths)

    # Set the credentials in the various sources
    for ctype, csource in sources:
        if csource == CredentialsSource.CONFIG:
            if ctype == CredentialsType.AUTH_TOKEN:
                config.api.auth_token = SecretStr(MOCK_TOKEN)
            elif ctype == CredentialsType.PASSWORD:
                config.api.username = MOCK_USER
                config.api.password = SecretStr(MOCK_PASSWORD)
        elif csource == CredentialsSource.ENV:
            if ctype == CredentialsType.AUTH_TOKEN:
                monkeypatch.setenv("ZABBIX_API_TOKEN", MOCK_TOKEN)
            elif ctype == CredentialsType.PASSWORD:
                monkeypatch.setenv("ZABBIX_USERNAME", MOCK_USER)
                monkeypatch.setenv("ZABBIX_PASSWORD", MOCK_PASSWORD)
        elif csource == CredentialsSource.FILE:
            if ctype == CredentialsType.SESSION:
                sidfile = SessionFile()
                sidfile.set_user_session(MOCK_URL, MOCK_USER, MOCK_TOKEN)
                sidfile.save(session_file)
                config.app.use_session_file = True
                config.app.allow_insecure_auth_file = True
            elif ctype == CredentialsType.AUTH_TOKEN:
                auth_token_file.write_text(f"{MOCK_USER}::{MOCK_TOKEN}")
                config.app.use_session_file = True
                config.app.allow_insecure_auth_file = True
            elif ctype == CredentialsType.PASSWORD:
                auth_file.write_text(f"{MOCK_USER}::{MOCK_PASSWORD}")
                config.app.auth_file = auth_file

    _, info = authenticator.login_with_any()

    # Provide more verbose assertion messages in case of failures
    # We want to know both source AND type when it fails.
    def fmt_assert_msg(
        expect_type: CredentialsType, expect_source: CredentialsSource
    ) -> str:
        return (
            f"Expected {expect_type} from {expect_source}, got {info.credentials.type} "
            f"from {info.credentials.source}"
        )

    assert info.credentials.type == expect_type, fmt_assert_msg(
        expect_type, expect_source
    )
    assert info.credentials.source == expect_source, fmt_assert_msg(
        expect_type, expect_source
    )

    assert info.token == MOCK_TOKEN

    # Ensure the login method modified the base renderable's zabbix version attribute
    # with the one we got from the mocked API response

    # Check the mocked version we replace the real class with
    assert table_renderable_mock.zabbix_version == Version("6.4.0")

    # Try to import the real class and check that our mock changes were applied
    from zabbix_cli.models import TableRenderable

    assert TableRenderable.zabbix_version == Version("6.4.0")

    # Check HTTP server errors
    httpserver.check_assertions()
    httpserver.check_handler_errors()


def test_table_renderable_mock_reverted() -> None:
    """Attempt to ensure that the TableRenderable has been unchanged after
     running tests that mutate it.

    If those tests have used the mock class correctly, this test should pass.

    TODO: Ensure this test is run _after_ the tests that mutate the class.
    """
    from zabbix_cli.models import TableRenderable

    # Ensure the mock changes were reverted
    assert TableRenderable.zabbix_version != Version("1.2.3")
    assert TableRenderable.zabbix_version.release == snapshot((7, 0, 0))


@pytest.fixture
def sample_session_file(tmp_path: Path) -> Path:
    """Create a sample session file for testing."""
    file_path = tmp_path / "sessions.json"
    file_path.write_text(
        '{"https://zabbix.example.com": [{"username": "admin", "session_id": "abc123"}]}'
    )
    return file_path


def test_sessionid_list_set_session() -> None:
    """Test setting sessions in SessionList."""
    session_list = SessionList()

    def check_session(username: str, session_id: str, expected_count: int):
        session = session_list.get_session(username)
        assert session is not None
        assert session.username == username
        assert session.session_id == session_id
        assert len(session_list) == expected_count

    assert len(session_list) == 0

    # Add new session
    session_list.set_session("user1", "abcd1234")
    check_session("user1", "abcd1234", 1)

    # Update the session
    session_list.set_session("user1", "xyz7890")
    check_session("user1", "xyz7890", 1)

    # Add another new session
    session_list.set_session("user2", "1234abcd")
    check_session("user2", "1234abcd", 2)


@pytest.mark.parametrize(
    "initial_sessions,username,expected",
    [
        ([], "user1", None),  # Empty list
        ([{"username": "user1", "session_id": "abc"}], "user1", "abc"),  # Existing user
        (
            [{"username": "user1", "session_id": "abc"}],
            "user2",
            None,
        ),  # Non-existent user
    ],
)
def test_sessionid_list_get_session(
    initial_sessions: list[dict[str, str]], username: str, expected: Optional[str]
):
    """Test retrieving sessions from SessionList."""
    session_list = SessionList(root=[SessionInfo(**s) for s in initial_sessions])
    result = session_list.get_session(username)

    if expected is None:
        assert result is None
    else:
        assert result is not None
        assert result.session_id == expected


@pytest.mark.parametrize(
    "url,username,expected_session",
    [
        ("https://zabbix1.com", "user1", None),  # Non-existent URL
        ("https://zabbix2.com", "user2", "xyz789"),  # Existing URL and user
    ],
)
def test_sessionid_file_get_user_session(
    url: str, username: str, expected_session: Optional[str]
):
    """Test retrieving user sessions from SessionFile."""
    session_file = SessionFile(
        root={
            "https://zabbix2.com": SessionList(
                root=[SessionInfo(username="user2", session_id="xyz789")]
            )
        }
    )

    result = session_file.get_user_session(url, username)
    if expected_session is None:
        assert result is None
    else:
        assert result is not None
        assert result.session_id == expected_session


@pytest.mark.parametrize(
    "url,username,session_id,expected_urls",
    [
        ("https://new.com", "user1", "session1", ["https://new.com"]),  # New URL
        (
            "https://existing.com",
            "user2",
            "session2",
            ["https://existing.com"],
        ),  # Existing URL
    ],
)
def test_sessionid_file_set_user_session(
    url: str, username: str, session_id: str, expected_urls: list[str]
):
    """Test setting user sessions in SessionFile."""
    session_file = SessionFile()
    session_file.set_user_session(url, username, session_id)

    assert set(session_file.root.keys()) == set(expected_urls)
    session = session_file.get_user_session(url, username)
    assert session is not None
    assert session.session_id == session_id


@pytest.mark.parametrize(
    "file_exists,secure_permissions,allow_insecure,should_raise",
    [
        (True, True, False, False),  # Normal case
        (False, True, False, True),  # File doesn't exist
        (True, False, False, True),  # Insecure permissions
        (True, False, True, False),  # Insecure permissions but allowed
    ],
)
def test_sessionid_file_load(
    tmp_path: Path,
    file_exists: bool,
    secure_permissions: bool,
    allow_insecure: bool,
    should_raise: bool,
):
    """Test loading SessionFile with various conditions."""
    file_path = tmp_path / "test_sessions.json"
    if file_exists:
        file_path.write_text('{"https://zabbix.example.com": []}')

    with patch("zabbix_cli.auth.file_has_secure_permissions") as mock_secure:
        mock_secure.return_value = secure_permissions

        if should_raise:
            with pytest.raises((SessionFileNotFoundError, SessionFilePermissionsError)):
                SessionFile.load(file_path, allow_insecure=allow_insecure)
        else:
            result = SessionFile.load(file_path, allow_insecure=allow_insecure)
            assert isinstance(result, SessionFile)
            assert result._path == file_path


@pytest.mark.parametrize(
    "has_path,allow_insecure,secure_permissions,should_raise",
    [
        (True, False, True, False),  # Normal case
        (False, False, True, True),  # No path specified
        (True, False, False, False),  # Insecure permissions, will be fixed
        (True, True, False, False),  # Insecure permissions allowed
    ],
)
def test_sessionid_file_save(
    tmp_path: Path,
    has_path: bool,
    allow_insecure: bool,
    secure_permissions: bool,
    should_raise: bool,
):
    """Test saving SessionFile with various conditions."""
    file_path = tmp_path / "test_sessions.json" if has_path else None
    session_file = SessionFile()
    if has_path:
        session_file._path = file_path

    with (
        patch("zabbix_cli.auth.file_has_secure_permissions") as mock_secure,
        patch("zabbix_cli.auth.set_file_secure_permissions") as mock_set_secure,
    ):
        mock_secure.return_value = secure_permissions

        if should_raise:
            with pytest.raises(SessionFileError) as exc_info:
                session_file.save(path=file_path, allow_insecure=allow_insecure)
            assert str(exc_info.value) == snapshot(
                "Cannot save session file without a path."
            )
        else:
            session_file.save(path=file_path, allow_insecure=allow_insecure)
            assert file_path and file_path.exists()
            if not secure_permissions and not allow_insecure:
                mock_set_secure.assert_called_once_with(file_path)



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import NamedTuple
from typing import Optional

from pydantic import BaseModel
from pydantic.fields import FieldInfo

from zabbix_cli.config.constants import CONFIG_PRIORITY
from zabbix_cli.config.constants import DEFAULT_CONFIG_FILE
from zabbix_cli.exceptions import ConfigError

if TYPE_CHECKING:
    from zabbix_cli.config.model import Config

logger = logging.getLogger(__name__)


def load_config_toml(filename: Path) -> dict[str, Any]:
    """Load a TOML configuration file."""
    try:
        import tomllib
    except ImportError:
        import tomli as tomllib

    try:
        return tomllib.loads(filename.read_text())
    except tomllib.TOMLDecodeError as e:
        raise ConfigError(f"Error decoding TOML file {filename}: {e}") from e
    except OSError as e:
        raise ConfigError(f"Error reading TOML file {filename}: {e}") from e


def load_config_conf(filename: Path) -> dict[str, Any]:
    """Load a conf configuration file with ConfigParser."""
    import configparser

    config = configparser.ConfigParser()
    try:
        config.read_file(filename.open())
        return {s: dict(config.items(s)) for s in config.sections()}
    except (configparser.Error, OSError) as e:
        raise ConfigError(
            f"Error reading legacy configuration file {filename}: {e}"
        ) from e


def find_config(
    filename: Optional[Path] = None,
    priority: tuple[Path, ...] = CONFIG_PRIORITY,
) -> Optional[Path]:
    """Find all available configuration files.

    :param filename: An optional user supplied file to throw into the mix
    """
    # FIXME: this is a mess.
    #        If we have a file, just try to load it and call it a day?
    filename_prio = list(priority)
    if filename:
        filename_prio.insert(
            0, filename
        )  # TODO: append last when we implement multi-file config merging
    for fp in filename_prio:
        if fp.exists():
            logging.debug("found config %r", fp)
            return fp
    return None


def get_config(filename: Optional[Path] = None, *, init: bool = False) -> Config:
    """Get a configuration object.

    Args:
        filename (Optional[str], optional): An optional user supplied file to throw into the mix. Defaults to None.

    Returns:
        Config: Config object loaded from file
    """
    from zabbix_cli.config.model import Config

    return Config.from_file(filename, init=init)


def get_replacement_fields(field: FieldInfo) -> list[str]:
    replacement: list[str] = []
    # NOTE: I think this is silly over doing `isinstance(field.json_schema_extra, dict)`
    #       but Pyright fails to infer types correctly when doing so, because it
    #       assumes the type is `dict[Unknown, Unknown]` after performing such
    #       isinstance checks. Not great.
    if isinstance(field.json_schema_extra, dict):
        # if field.json_schema_extra and not callable(field.json_schema_extra):
        rep = field.json_schema_extra.get("replacement")
        if isinstance(rep, str):
            replacement.append(rep)
        elif isinstance(rep, list):
            replacement = list({r for r in rep if isinstance(r, str)})
    return replacement


# TODO: can we bake this into get_deprecated_fields_set? Should we?
def check_deprecated_fields(model: BaseModel) -> None:
    """Check for deprecated fields in a model and log a warning."""
    # Sort for reproducibility + readability
    for field_name in sorted(model.model_fields_set):
        f = model.model_fields.get(field_name)
        if not f:
            continue
        if f.deprecated:
            if replacements := get_replacement_fields(f):
                from zabbix_cli.output.console import warning

                r = ", ".join(replacements)
                warning(
                    f"Config option [configopt]{field_name}[/] is deprecated. Replaced by: [configopt]{r}[/]."
                )
            else:
                logger.warning("Config option `%s` is deprecated.", field_name)


def get_deprecated_fields_set(
    model: BaseModel, parent: Optional[str] = None
) -> list[DeprecatedField]:
    """Get a list of deprecated fields set on a model and all its submodels."""
    fields: list[DeprecatedField] = []
    # Sort for reproducibility + readability
    for field_name in sorted(model.model_fields_set):
        field = model.model_fields.get(field_name)
        if not field:
            continue

        # Get field value safely
        try:
            value = getattr(model, field_name)
        except AttributeError:
            logger.error(
                "Field %s.%s exists in model_fields_set but is not accessible",
                model,
                field_name,
            )
            continue

        # Recurse into submodels
        if isinstance(value, BaseModel):
            submodel_fields = get_deprecated_fields_set(value, parent=field_name)
            fields.extend(submodel_fields)
        else:
            # We have a field that is not a submodel
            if not field.deprecated:
                continue
            name = f"{parent}.{field_name}" if parent else field_name
            replacements = get_replacement_fields(field)
            fields.append(DeprecatedField(name, value, replacements))
    return fields


def replace_deprecated_fields(model: BaseModel) -> None:
    """Set replacement fields for deprecated fields in a model.

    I.e. if field `old_field` is deprecated and replaced by `new_field`,
    set `new_field` to the value of `old_field` if `new_field` is not already set.
    """
    deprecated_fields = get_deprecated_fields_set(model)
    for field in deprecated_fields:
        if not field.replacement:
            continue
        # Update the model with the new field
        for replacement in field.replacement:
            _set_replacement_field(model, field, replacement)


def _set_replacement_field(
    model: BaseModel, field: DeprecatedField, replacement: str
) -> None:
    try:
        # Decompose the replacement field into its attributes
        attributes = replacement.split(".")
        to_replace = model
        for attr in attributes[:-1]:
            to_replace = getattr(to_replace, attr)
        field_to_update = attributes[-1]

        # Don't update if replacement field is already set
        if (
            isinstance(to_replace, BaseModel)
            and field_to_update in to_replace.model_fields_set
        ):
            logger.debug("Field `%s` is already set, skipping", field.replacement)
            return

        setattr(to_replace, field_to_update, field.value)
    except AttributeError as e:
        logger.error(
            "Failed to update field `%s` with value `%s` from  deprecated field `%s`: %s",
            field.replacement,
            field.value,
            field.field_name,
            e,
        )


def fmt_deprecated_fields(fields: list[DeprecatedField]) -> str:
    """Format a list of deprecated fields for logging."""
    lines: list[str] = []
    for field in fields:
        if field.replacement:
            replacements = ", ".join(f"[option]{r}[/]" for r in field.replacement)
            lines.append(
                f"  - [option]{field.field_name}[/] (replaced by: {replacements})"
            )
        else:
            lines.append(f"  - [option]{field.field_name}[/] (removed)")
    return "\n".join(lines)


class DeprecatedField(NamedTuple):
    """A deprecated field in a model."""

    field_name: str
    value: Any
    replacement: list[str] = []


def init_config(
    config: Optional[Config] = None,
    config_file: Optional[Path] = None,
    *,
    overwrite: bool = False,
    # Compatibility with V2 zabbix-cli-init args
    url: Optional[str] = None,
    username: Optional[str] = None,
) -> Config:
    """Creates required directories and boostraps config with
    options required to connect to the Zabbix API.
    """

    from zabbix_cli.config.model import Config
    from zabbix_cli.dirs import init_directories

    # Create required directories
    init_directories()

    if config_file is None:
        config_file = DEFAULT_CONFIG_FILE
    if config_file.exists() and not overwrite:
        raise ConfigError(
            f"File {config_file} already exists. Use [option]--overwrite[/] to overwrite it."
        )

    if not config:
        config = Config.sample_config()
    config.config_path = config_file

    if url:
        config.api.url = url
    if username:
        config.api.username = username

    return config



"""This module is based on PyZabbix (https://github.com/lukecyca/pyzabbix),
which is licensed under the GNU Lesser General Public License (LGPL) according
to its PyPI metadata.

It is unclear which version of PyZabbix was vendored into Zabbix-CLI originally,
but we can assume it's not a version later than 0.7.4, which was the last version
available on PyPI before the majority of the code was vendored, as evidenced by
this git blame: https://github.com/unioslo/zabbix-cli/blame/2.3.2/zabbix_cli/pyzabbix.py

We assume that the copyright years of the original PyZabbix code are from 2013-2015
for that reason. The source code repository contains no LICENSE file, even
though its metadata states that it is LGPL-licensed.

An abbreviated version of the LGPL-3.0 license text is included below:

    Copyright (C) 2013-2015 PyZabbix Contributors
    Modified work Copyright (C) 2022-2024 University of Oslo

    This library is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library. If not, see <https://www.gnu.org/licenses/>.

Additional Notices:
- This code was originally vendored by Zabbix-CLI from PyZabbix
- Modifications have been made to the original PyZabbix code to adapt it
  for use in this project. It is _very_ different, and it's unclear if
  we should even call it PyZabbix anymore. It's more like a fork.
- The original source code can be found at: https://github.com/lukecyca/pyzabbix
"""

from __future__ import annotations



from __future__ import annotations

from pydantic import ByteSize

from .constants import NONE_STR


def bytesize_str(b: int | None, *, decimal: bool = False) -> str:
    if b is None or b < 0:
        return NONE_STR
    return ByteSize(b).human_readable(decimal=decimal)



"""This module is based on click-repl (https://github.com/click-contrib/click-repl) @ f08ba39

Click-repl is licensed under the MIT license and has the following copyright notices:
    Copyright (c) 2014-2015 Markus Unterwaditzer & contributors.
    Copyright (c) 2016-2026 Asif Saif Uddin & contributors.

    Permission is hereby granted, free of charge, to any person obtaining a copy of
    this software and associated documentation files (the "Software"), to deal in
    the Software without restriction, including without limitation the rights to
    use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

The module uses the original click-repl code as a basis for providing a REPL
with autocompletion for the application. Several modifications have been made
to adapt the code to the needs of Zabbix-CLI, as well as removing compatibility
with Python 2 and Click < 8.0.

Most prominently, the code has been refactored to take in a Typer app when
starting the REPL, which allows us to pass in more information about the available
commands and options to the REPL.

Furthermore, type annotations have been added to the entire vendored codebase,
and changes have been made in order to make the original code pass type checking.
Among these changes is removing code that adds compatibility with Click < 8.0, which
relied a lot on duck typing and was generally not type-safe.
"""
